#include "defns.h"
#include "types.h"
#include "extern.h"

float* BoostingExampleWeights;
float* BoostingClassifierWeights;

void init_boosting()
{
  int i;

  BoostingExampleWeights=malloc(number_of_examples*sizeof(float));
  BoostingClassifierWeights=malloc(number_of_trees*sizeof(float));
  for(i=0;i<number_of_examples;i++)
    BoostingExampleWeights[i]=1.0/(float)number_of_examples;
}

// AFTER THE BOOSTING ALGORITHM HAS BEEN RUN, THE WEIGHTS OF THE VARIOUS
// DATA EXAMPLES MUST BE ADJUSTED.  IN ADDITION, THE CLASSIFIER MUST BE
// ASSIGNED A WEIGHT VALUE BASED ON ITS ERROR.  THIS FUNCTION PERFORMS
// BOTH OF THOSE FUNCTIONS
//
//   INPUT:  (1) THE COMPLETE TRAINING SET
//           (2) THE TREE CREATED FROM BOOSTING
//
//  OUTPUT:  (1) THE WEIGHT OF THE BOOSTED CLASSIFIER
//
float UpdateWeights(Tree myTree)
{
  int x;
  float Correct, mySum, myEpsilon;

  Correct=0;
  for(x=0;x<number_of_examples;x++)
    if (Category(train[x],myTree)==Class(train[x]))
      Correct+=BoostingExampleWeights[x];
  myEpsilon=1.0-Correct;

  if (myEpsilon >= 1.0-(1.0/(float)number_of_classes)) return -1;
  if (myEpsilon == 0)                       return -2;

  mySum=0;
  for(x=0;x<number_of_examples;x++)
  {
    if (Category(train[x],myTree)==Class(train[x]))
      BoostingExampleWeights[x]*= 1.0/(2.0*(1.0-myEpsilon));
    else
      BoostingExampleWeights[x]*= 1.0/(2.0*myEpsilon);
    mySum += BoostingExampleWeights[x];
  }
  for(x=0;x<number_of_examples;x++)
    BoostingExampleWeights[x] /= mySum;
  return log(  ( (number_of_classes-1)*(1.0-myEpsilon) )  /myEpsilon);
}

int BoostingUpdateWeights(int i)
{
  if (use_pruning)
    BoostingClassifierWeights[i]=UpdateWeights(pruned_trees[i]);
  else
    BoostingClassifierWeights[i]=UpdateWeights(unpruned_trees[i]);
    
  if (BoostingClassifierWeights[i]==-1)  return -1;  /* Special code for error too great, delete classifier */
  if (BoostingClassifierWeights[i]==-2)  return -2;  /* Special code for no error, stop building */

  return 0;
}
