/*************************************************************************/
/*									 */
/*	Main routine, c4.5						 */
/*	------------------						 */
/*									 */
/*************************************************************************/
#include <sys/types.h>
#include <unistd.h>
#include <time.h>
#ifndef _GNU_SOURCE
  #include "getopt.h"
#else
  #include <getopt.h>
#endif
#ifdef Linux
 #include <fpu_control.h>
 fpu_control_t fpu_current_state=0;
#endif
#ifdef Linux
int setfpu()
{
  fpu_control_t t;

  _FPU_GETCW(fpu_current_state);
  t=(fpu_current_state&0xfcff) | 0x0000;
  _FPU_SETCW(t);
  return 0;
}
int restorefpu()
{
   return 0;
   _FPU_SETCW(fpu_current_state);
   return 0;
}
#endif
#ifdef USE_MPI
   #include <mpi.h>
#endif
#include "defns.h"
#include "types.h"
#include "extern.h"





/*  External data, described in extern.i  */

int		MaxAtt, MaxClass, MaxDiscrVal = 2;
int MaxItem;
int	*MaxAttVal;
char            **class_name,
		**attribute_name,
		***AttValName;

int		VERBOSITY = 0;
int		TRIALS    = 1;

/* Many Boolean flags. Descriptions to be included enventually. */
Boolean  GAINRATIO  = True;
Boolean  SUBSET     = False;
Boolean  PROBTHRESH = False;
Boolean  RESUBSTITUTION=False;
Boolean  USE_HOEFFDING_PARALLEL_TREES=False;
Boolean  SAVE_PRUNED_TREES=True;
Boolean  SAVE_UNPRUNED_TREES=True;
Boolean  LOAD_PRUNED_TREES=False;
Boolean  LOAD_UNPRUNED_TREES=False;
Boolean  LOAD_TEST_SET=False;
Boolean  LOAD_TRAIN_SET=True;
Boolean  PRINT_ALL_TREES=False;
Boolean  PRUNING_REQUESTED=True;

int		MINOBJS   = 2,
		WINDOW    = 0,
		INCREMENT = 0;






Boolean		AllKnown = True;
int best_tree=0;
int tree_stats=0;



/* Administrative variables */
int my_mpi_rank=0;
char *progname;    
char version[]="USFC4.5 R8  $Revision: 2.15 $";


/* Local functions */
int _process_options();


int main(int argc, char **argv)
{

  #ifdef USE_MPI
     MPI_Init(&argc, &argv);
     MPI_Comm_rank(MPI_COMM_WORLD,&my_mpi_rank);
  #endif

  progname=argv[0];
  

  /* Set any default values, then we call process_options */
  _process_options(argc,argv);


  /* Included to set the floating-point unit on x86 systems */
#ifdef Linux
  setfpu();
#endif

  /* We optionally load the training data and test data */
  load_data(LOAD_TRAIN_SET, LOAD_TEST_SET);


  /* Possibly load stored trees */
  if ( LOAD_UNPRUNED_TREES ) unpruned_trees=load_unpruned_trees();
  if ( LOAD_PRUNED_TREES   ) pruned_trees=load_pruned_trees();

  /* Create the trees, if necessary. Pruning is also done here. */
  create_trees();


  /*  Soften thresholds in best tree  */
  if ( PROBTHRESH ) {
    printf("Softening thresholds");
    if ( use_windowing ) printf(" for best tree from trial %d", best_tree);
    printf("\n");
    SoftenThresh(pruned_trees[best_tree]);
    printf("\n");
    print_tree(pruned_trees[best_tree]);
  }

  /* Save resulting trees */
  if ( SAVE_UNPRUNED_TREES && ! LOAD_UNPRUNED_TREES && unpruned_trees != NULL )
  {
    save_unpruned_trees(unpruned_trees);
  }
  if ( SAVE_PRUNED_TREES && ! LOAD_PRUNED_TREES && pruned_trees != NULL )
  {
    save_pruned_trees(pruned_trees);
  }

  /* Divide and conquer method */
  if ( divide_and_conquer )  {
    /* Divide the training data */
    divide_train(use_pruning);
    
    /* Then the test data */
    divide_test(use_pruning);
  }
  


  /* Possibly print out confusion matrix for training set */
  if ( evaluate_training_set ) {
    /*  Evaluation  */
    load_training_data();
    printf("\n\nEvaluation on training data (%d items):\n", 
	   number_of_train_examples);
    evaluate_tree(True, best_tree);
  }
  if ( evaluate_test_set ) {   
    load_test_data();
    printf("\nEvaluation on test data (%d items):\n",number_of_test_examples);
    evaluate_tree(True, best_tree);
  }
  
  /* Output predictions to file. This uses either the train or test set */
  if ( do_predictions ) 
    predict(RESUBSTITUTION);

  
#ifdef Linux
  restorefpu();
#endif 

#ifdef USE_MPI
   MPI_Finalize();
#endif

  return 0;
}









extern char *optarg;
extern int optind;
char options[]="hf:bupv::t:w:i:gsm:c:PaUrIQDR:X:M:N:F:A:LTB::Y::V::O";

/* Long options. This gives us enough room to add as many new features
   as we want without name collisions. */
enum {no_option, option_version,option_help,option_useunprunedtree,
      option_predict,option_hoeffdingparalleltrees,option_randtreemaxdepth,option_savetrees,option_nosavetrees,option_mpi,option_reduced_error_pruning,option_load_validation_set,option_stats
};
struct option longopts[]=
{
  {"version",           no_argument,       0, option_version},
  {"help",              no_argument,       0, 'h'},

  /* Initial C4.5 options, extended to long options */
  {"filestem",          required_argument, 0, 'f'},
  {"batch",             no_argument,       0, 'b'},
  {"test",              no_argument,       0, 'u'},
  {"verbosity",         optional_argument, 0, 'v'},
  {"trials",            required_argument, 0, 't'},
  {"window",            required_argument, 0, 'w'},
  {"window-incr",       required_argument, 0, 'i'},
  {"use-gain",          no_argument,       0, 'g'},
  {"min-objs",          required_argument, 0, 'm'},
  {"confidence",        required_argument, 0, 'c'},
  {"use-subsetting",    no_argument,       0, 's'},

  /* Add new ones here */
  {"stats", no_argument,                 0, option_stats},
  {"no-pruning",      no_argument,       0, 'Q'},
  {"predict-all-trees",      no_argument,       0, 'a'},
  {"use-unpruned-tree",      no_argument,       0, 'U'},
  {"use-validation-set",     optional_argument,0, 'V'},
  {"load-validation-set",    no_argument,0,option_load_validation_set},
  {"predict",                no_argument,       0, 'P'},
  {"resubstitution",         no_argument,       0, 'r'},
  {"hoeffding-parallel-trees", no_argument,     0,
   option_hoeffdingparalleltrees},
  {"use-index-file", no_argument,              0, 'I'},
  {"delete-trees", no_argument,                0, 'D'},
  {"random-split", required_argument,          0, 'R'},
  {"random-forest", required_argument,         0, 'F'},
  {"random-subspaces", required_argument,       0, 'S'},
  {"random-attributes", required_argument,      0, 'A'},
  {"bagging", optional_argument,               0, 'B'},
  {"seed", required_argument,                  0, 'X'},
  {"depth", required_argument,                 0, 'M'},
  {"number-of-trees",required_argument,        0, 'N'},
  {"load-trees",no_argument,                   0, 'T'},
  {"load-pruned-trees",no_argument,(int *)&LOAD_PRUNED_TREES,True},
  {"load-pruned-tree", no_argument,(int *)&LOAD_PRUNED_TREES,True},
  {"load-unpruned-trees",no_argument,(int *)&LOAD_UNPRUNED_TREES,True},
  {"load-unpruned-tree",no_argument,(int *)&LOAD_UNPRUNED_TREES,True},
  {"no-load-pruned-trees",no_argument,(int *)&LOAD_PRUNED_TREES,False},
  {"no-load-pruned-tree", no_argument,(int *)&LOAD_PRUNED_TREES,False},
  {"no-load-unpruned-trees",no_argument,(int *)&LOAD_UNPRUNED_TREES,False},
  {"no-load-unpruned-tree",no_argument,(int *)&LOAD_UNPRUNED_TREES,False},
  {"mpi", required_argument,0,option_mpi},
  {"random-tree-max-depth", required_argument, 0,option_randtreemaxdepth},
  {"collapse", no_argument,         (int *)&collapse_tree, True},
  {"no-collapse", no_argument,      (int *)&collapse_tree, False},
  {"save",        no_argument,      0, option_savetrees},
  {"no-save",     no_argument,      0, option_nosavetrees},
  {"conquer", no_argument,          (int *)&divide_and_conquer, True},
  {"no-conquer", no_argument,       (int *)&divide_and_conquer, False}, 
  {"use-example-weights", no_argument, (int *)&use_example_weights, True},
  {"no-use-example-weights", no_argument, (int *)&use_example_weights,False},
  {"hard-limit-min-branch",no_argument, (int *)&hard_limit_min_branch,True},
  {"no-hard-limit-min-branch",no_argument, (int *)&hard_limit_min_branch,False},
  {"evaluate-training-set",no_argument, (int *)&evaluate_training_set, True},
  {"no-evaluate-training-set",no_argument, (int *)&evaluate_training_set,False},
  {"load-training-set",no_argument, (int *)&LOAD_TRAIN_SET, True},
  {"no-load-training-set", no_argument, (int *)&LOAD_TRAIN_SET, False},
  {"reduced-error-pruning", no_argument, 0, option_reduced_error_pruning},
  {"subtree-raising",no_argument, (int *)&use_subtree_raising, True},
  {"no-subtree-raising", no_argument, (int *)&use_subtree_raising, False},
  {"error-estimate",no_argument, (int *)&use_error_estimate, True},
  {"no-error-estimate", no_argument, (int *)&use_error_estimate, False},
  {"use-oob-estimate", no_argument, 0, 'O'},
  {"use-gini",no_argument,(int *)&use_gini, True},
  {"no-use-gini", no_argument,(int *)&use_gini, False},
  {"boosting", optional_argument, 0, 'Y'},
  {"pure",optional_argument,0,'Z'}, /*build pure trees*/
  {"leafs",no_argument,0, 'L'},
  /* Last element must be all zeroes */
  {0,0,0,0}
};
char usage[]="\n"
"Usage: %s -f filestem\n"
"  Original C4.5 Options\n"
"   -h | --help               Print this screen\n"
"   --version                 Print version information and exit\n"
"\n"
"   -f s | --filestem s       Use s as filestem\n"
"\n"
"   Original options\n"
"   ================\n"
"   -b   | --batch            Use batch mode\n"
"   -u   | --test             Test trees against test set filestem.test\n"
"   -p                        Use probabilistic estimates\n"
"   -v n | --verbosity n      Use verbosity level n\n"
"   -t n | --trials n         Run n trials\n"
"   -w n | --window n         Set initial window size\n"
"   -i n | --window-incr n    Use n as max. number of objects added at\n"
"                             each iteration to window\n"
"   -s   | --use-subsetting   Force subsetting on discrete attributes\n"
"                             with more than two values\n"
"\n"
"   Misc options\n"
"   ============\n"
"   --conquer		     	Save data at each leaf for a divide and conquer\n"
"      --no-conquer           	algorithm.\n"
"   --use-example-weights     	Read in example weights from .weights file\n"
"      --no-use-example-weights\n"
"\n"
"   File IO Options\n"
"   ===============\n"
"   -P | --predict            	Output predictions using best tree to \n"
"                             	filestem.pred\n"
"   -a | --predict-all-trees     Save .pred files for all trees predicted\n"
"                                Used in addition to the -P (--predict) flag\n"
"   -r | --resubstitution     	Print confusion matrix for training set\n"
"   -I | --use-index-files    	Use index file format to read in data.\n"
"   -D | --delete-trees       	Do not save .tree and .unpruned\n"
"   -T | --load-trees            Load both pruned/unpruned trees from files\n"
"   -S | --stats                Output tree height, number of nodes and leafs\n"
"   --save                    	Save files (including tree) (default)\n"
"      --no-save\n"
"   -V[=n] |\n"
"     --use-validation-set[=n]   Use n% of training set as validation data\n"
"                                  for pruning.\n"
"   --load-validation-set	Load the validation set from filestem.validation\n"
"   --evaluate-training-set   	Run the evaluation (including cm)\n"
"				on the train set\n"
"      --no-evaluate-training-set\n"
"   --load-pruned-tree[s]        Load pruned tree(s) from .pruned_tree\n"
"      --no-load-pruned-tree[s]\n"
"   --load-unpruned-tree[s]      Load unpruned tree(s) from .unpruned_tree\n"
"      --no-load-unpruned-tree[s]\n"
"\n"
"\n"
"   Randomization Options \n"
"   =====================\n"
"   -R n | --random-split n   	Randomly choose from the top n splits at a node\n"
"				 (Use -1 for all attributes)\n"
"   --random-tree-max-depth=n 	Use n as the maximum depth from which to choose\n"
"			     	random splits.\n"
"   -F n | --random-forest=n  	Create splits from subset of only n attributes\n"
"				Use 0 to use Breiman heuristic (sqrt(s))\n"
"   -S n | --random-subspaces=n 	Use a random subset of features to build tree\n"
"   -A n | --random-attributes=n Like random split, but choose from the top\n"
"				n attributes (not splits). Use -1 for all attrs.\n"
"   -X n | --seed n           	Use n as the random seed\n"
"\n"
"\n"
"   -N n | --number-of-trees=n 	Create n trees and vote them\n"
"   -B[=n] | --bagging[=n]       Use bagging at n% of the data (use with\n"
"                                  --number-of-trees). Default is 100%.\n"
"   -O     | --use-oob-estimate  Use out-of-bag estimates\n"
"   -Y     | --boosting[=n]      Use boosting. If n=0 then boost the\n"
"                                pruned trees, else boost the unpruned\n"
"\n"
"   Tree Growing \n"
"   ====================\n"
"   -M n | --depth n	     	Limit tree to a depth of n\n"
"   --collapse		     	Collapse trees (default)\n"
"      --no-collapse        \n"
"   --hard-limit-min-branch   	Impose a hard limit on the min. number of \n"
"				examples at a split\n"
"      --no-hard-limit-min-branch  \n"
"   -g   | --use-gain            Use GainRatio instead of Gain\n"
"\n"
"   --use-gini                   Use the gini index\n"
"      --no-use-gini\n"
"  --pure                       Build pure trees\n"
"\n"
"   Pruning Flags\n"
"   =============\n"
"   -U | --use-unpruned-tree  	When predictions are output, use unpruned tree\n"
"   -Q | --no-pruning            Disable pruning entirely\n"
"   -m n | --min-objs n          In all tests, at least two branches must\n"
"                                contain a minimum of n objects to split.\n"
"   -c n | --confidence n        Use n%% as confidence level\n"
"\n"
"\n"
"   Options for Reduced-Error Pruning \n"
"   =================================\n"
"   --subtree-raising            When pruning, replace a node with the best\n"
"      --no-subtree-raising        sub-branch if needed.\n"
"   --error-estimate		Estimate the error using EBP (error-based\n"
"      --no-error-estimate	  pruning).\n"
"   -V[=n] |\n"
"     --use-validation-set[=n]   Use n%% of training set as validation data\n"
"                                  for pruning.\n"
"\n"
"   --reduced-error-pruning      Use reduced error pruning on the dataset.\n"
"				 Turns off subtree raising, error estimates,\n"
"				 and uses a separate validation set.\n"
"\n";







int _process_options(int argc, char **argv)
{
  int o;
  long seed=0;
  int option_index;
  char *old_filestem;
  


  PrintHeader("decision tree generator");
  
  /*  Process options  */
  while ( (o = getopt_long(argc, argv, options,longopts,&option_index)) != EOF ) {
    switch (o) {
    case 'L':
      OUTPUT_LEAFS = True;
      printf("\tOutput Test set Leafs\n");
      break;
    case 0:
      printf("\toption %s set\n", longopts[option_index].name);
      break;
    case 'f':   
      filestem = strdup(optarg);
      printf("\tFile stem <%s>\n", filestem);
      break;
    case 'b':   
      use_windowing=False;
      printf("\tWindowing disabled (now the default)\n");
      break;
    case 'u':   
      evaluate_test_set = True;
      LOAD_TEST_SET=True;
      printf("\tTrees evaluated on unseen cases\n");
      break;
    case 'p':
      PROBTHRESH = True;
      printf("\tProbability thresholds used\n");
      break;
    case 'v':   
      if ( optarg ) VERBOSITY = atoi(optarg);
      else          VERBOSITY = 1;
      printf("\tVerbosity level %d\n", VERBOSITY);
      break;
    case 't':   
      TRIALS = atoi(optarg);
      printf("\tWindowing enabled with %d trials\n", TRIALS);
      if ( TRIALS < 1 || TRIALS > 10000 )
         die("\t** illegal value **\n");
      use_windowing=True;
      break;
    case 'w':   
      WINDOW = atoi(optarg);
      printf("\tInitial window size of %d items\n", WINDOW);
      if ( WINDOW < 1 || WINDOW > 1000000 )
         die("\t** illegal value **\n");
      use_windowing=True;
      break;
    case 'i':   
      INCREMENT = atoi(optarg);
      printf("\tMaximum window increment of %d items\n", INCREMENT);
      if ( INCREMENT < 1 || INCREMENT > 1000000 )
         die("\t** illegal value **\n");
      use_windowing=True;
      break;
    case 'g':   
      GAINRATIO = False;
      printf("\tGain criterion used\n");
      break;
    case 's':   
      SUBSET = True;
      printf("\tTests on discrete attribute groups\n");
      break;
    case 'm':   
      MINOBJS = atoi(optarg);
      printf("\tSensible test requires 2 branches with >=%d cases\n",MINOBJS);
      if ( MINOBJS < 1 || MINOBJS > 1000000 )
         die("\t** illegal value **\n");
      break;
    case 'c':   
      confidence_factor = atof(optarg);
      printf("\tPruning confidence level %g%%\n", confidence_factor);
      if ( confidence_factor < Epsilon || confidence_factor > 100 )
         die("\t** illegal value **\n");
      confidence_factor /= 100;
      break;
    case 'U':
      use_pruning=False;
      printf("\tUsing unpruned tree for evaluation\n");
      break;
    case 'P':
      LOAD_TEST_SET=True;
      do_predictions=True;
      printf("\tOutput predictions\n");
      break;
    case 'a':
      PRINT_ALL_TREES=True;
      printf("\tSaving predictions on all trees\n");
      break;
    case 'r':
      RESUBSTITUTION=True;
      break;
    case option_hoeffdingparalleltrees:
      USE_HOEFFDING_PARALLEL_TREES=True;
      printf("\tUsing Hoeffding parallel trees\n");
      break;
    case 'I':
      use_index_files=True;
      printf("\tUsing index file format\n");
      break;
    case 'D':
      SAVE_PRUNED_TREES=False;
      SAVE_UNPRUNED_TREES=False;
      printf("\tTrees will not be saved\n");
      break;
    case option_savetrees:
      SAVE_PRUNED_TREES=True;
      SAVE_UNPRUNED_TREES=True;
      printf("\tTrees will be saved\n");
      break;
    case option_nosavetrees:
      SAVE_PRUNED_TREES=False;
      SAVE_UNPRUNED_TREES=False;
      printf("\tTrees will not be saved\n");
      break;
    case 'R':
      rand_split_max = atoi(optarg);
      printf("\tPick from %d best splits\n", rand_split_max);
      break;
    case option_randtreemaxdepth:
      max_random_tree_depth=atoi(optarg);
      printf("\tLimit randomization to first %d levels\n",max_random_tree_depth);
      break;
    case 'X':
      seed=atol(optarg);
      printf("\tUsing %ld as the seed\n", seed);
      break;
    case 'M':
      max_tree_depth=atoi(optarg);
      printf("\tUsing %d as max tree depth\n", max_tree_depth);
      break;
    case 'N':
      number_of_trees=atoi(optarg);
      printf("\tCreating %d total trees\n", number_of_trees);
      break;
    case 'F':
      use_random_forest=True;
      number_of_random_attributes=atoi(optarg);
      use_bagging=True;
      PRUNING_REQUESTED=False;
      use_pruning=False;
      printf("\tUsing random forests (subsets of %d)\n", number_of_random_attributes );
      printf("\tUsing bagging\n");
      printf("\tDisabled pruning\n");
      printf("\tDisabled testing on pruned trees\n");
      break;
    case 'A':
      use_random_attribute_split=True;
      rand_split_max=atoi(optarg);
      printf("\tUsing random attribute splits (%d attrs)\n",rand_split_max);
      break;
    case 'S':
      use_random_subspaces=True;
      number_of_random_attributes=atoi(optarg);
      printf("\tUsing random subspaces (subsets of %d)\n", number_of_random_attributes);
      break;
    case 'T':
      LOAD_PRUNED_TREES=True;
      LOAD_UNPRUNED_TREES=True;
      printf("\tLoading tree files\n");
      break;
    case 'V':
      use_validation_set=True;
      if ( optarg ) validation_percent=atoi(optarg);
      printf("\tValidation set of size %d%% being used.\n",validation_percent);
      break;
    case option_load_validation_set:
      use_validation_set=True;
      load_validation=True;
      printf("\tLoading validation set from file\n");
      break;
    case 'B':
      use_bagging=True;
      if ( optarg ) bagging_percent=atoi(optarg);
      printf("\tBagging with %d%% of data\n", bagging_percent);
      break;
    case 'Q':
      PRUNING_REQUESTED=False;
      printf("\tPruning has been disabled\n");
      break;
    case option_reduced_error_pruning:
      use_validation_set=True;
      use_subtree_raising=False;
      use_error_estimate=False;
      printf("\tUsing REP with validation set, no subtree raising and no error estimates\n");
      break;
    case 'O':
      use_oob=True;
      printf("\tUsing OOB estimates\n");
      break;
    case option_mpi:
      my_mpi_rank+=atoi(optarg);
      break;
    case 'Y':
      use_boosting=True;
      printf("\tBoosting\n");
      break;
    case 'Z':
      collapse_tree=False;
      MINOBJS = 1;
      PURE_TREES = True;
      printf("\tPure Trees\n");
      break;
    case option_stats:
      tree_stats=1;
      break;
    case '?':   printf("unrecognised option: %c,\n", (char) o);
    case option_version:
      fprintf(stdout,"%s\n",version);
      exit(1);
    case 'h':case option_help:
      fprintf(stdout,usage,progname);
      exit(1);
    }
  }
#ifdef USE_MPI
  old_filestem=filestem;
  filestem=calloc(strlen(old_filestem)+my_mpi_rank/10+5,sizeof(char));
  sprintf(filestem,"%s-%d", old_filestem, my_mpi_rank);
  printf("\tUsing File stem <%s>\n", filestem);
  free(old_filestem);
#endif
 

  if ( divide_and_conquer ) {
     LOAD_TEST_SET=True;
     use_pruning=False;
  }

  /* 
     Initialize random number generator. If seed is set to 0 (default),
     use time otherwise use the seed.
  */
  if ( seed == 0 ) {
    seed = time(NULL)+getpid();
  } 
  seed+=my_mpi_rank;
  srand48(seed);
  printf("\tRandom seed is %ld\n", seed);
  



  return 0;
}







