/*
 *
 * Software designed by David Cieslak
 * This generates Decision Trees based on the C4.5 algorithm
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <float.h>
#include "globals.h"
#include "getopt.h"
#include "utils.h"

#ifdef Linux
 #include <fpu_control.h>
 fpu_control_t fpu_current_state=0;
#endif
#ifdef Linux
int setfpu()
{
  fpu_control_t t;

  _FPU_GETCW(fpu_current_state);
  t=(fpu_current_state&0xfcff) | 0x0000;
  _FPU_SETCW(t);
  return 0;
}
int restorefpu()
{
   return 0;
   _FPU_SETCW(fpu_current_state);
   return 0;
}
#endif

void print_names();
void print_data();
void print_test();
void print_tree();
void print_splits();
void print_level(struct tree_node * t, int level);
void print_level_cont(struct tree_node * t, int level);
int _process_options(int argc, char **argv);


float TotalInfo(float V[], int MinVal, int MaxVal);

int main(int argc, char *argv[])
{

#ifdef Linux
  setfpu();
#endif

  //sprintf(filestem, "%s", argv[1]);
  //filestem=strdup(argv[1]);
  _process_options(argc, argv);

  GetNames();
  printf("got names\n");
  GetData(DATA);
  printf("got data\n");
  
  create_tree();
  printf("created tree\n");
  print_tree();

  /* no testing set yet  */
  if(LOAD_TEST_SET==1) {
    GetData(TEST);

    if(do_predictions==1) {
      evaluate(TEST);
      print_class(TEST);
    }
  }
  /* make up some evaluation function here! */

#ifdef Linux
  restorefpu();
#endif


  return 0;
}

void print_names()
{
  int i,j;

  printf("Read .names file successfully\n");
  printf("\tclasses: ");
  for(i=0;i<num_classes;i++) {
    printf("%s ", classes_list[i]);
  }
  printf("\n");
  for(i=0;i<num_attrs;i++) {
    printf("\tattribute: %s\n",attr_name_list[i]);
      
    for(j=0;j<num_attr_vals[i];j++) {
      printf("\t\t%s\n",attr_list[i][j]);
    }
    printf("\n");
  }
  printf("\n");
}

void print_data()
{
  int i,j;

  printf("Read .data file successfully\n");
  for(i=0;i<num_train;i++) {
    printf("X%d = ",i);
    for(j=0;j<num_attrs;j++) {
      if(attr_types[j]==NOMINAL) {
	printf("%s,",attr_list[j][train_attr_list[i][j]]);
      }
      else {
	printf("%f,",train_continuous_list[i][j]);
      }
    }
    printf("Y%d = %s\n",i, classes_list[train_class_list[i]] );
  }
  printf("\n");
}

void print_test()
{
  int i,j;

  printf("Read .test file successfully\n");
  for(i=0;i<num_test;i++)
    {
      printf("X%d = ",i);
      for(j=0;j<num_attrs;j++)
	{
	  printf("%s ",attr_list[j][test_attr_list[i][j]]);
	}
      printf("Y%d = %s\n",i, classes_list[test_class_list[i]] );
    }
  printf("\n");
}

void print_tree()
{
  int i,j;
  printf("formed tree successfully\n");
  printf("\n\nPRINTING TREE:\n");
  printf("==============\n");
  for(i=0;i<tree_root->num_children;i++) {
      //printf("child: %d type: %d\n",i,tree_root->attribute);
      if(attr_types[tree_root->attribute]==CONTINUOUS) {
	if(i==0) {
	  printf("%s <= %f ",attr_name_list[tree_root->attribute],tree_root->split_val);
          if(thain==1) {
             printf("( ");
             for(j=0;j<num_classes;j++) {
                printf("%s:%d ",classes_list[j],tree_root->children[i]->class_dist[j]);
             }
             printf(")");
          }
	}
	else {
	  printf("%s > %f ",attr_name_list[tree_root->attribute],tree_root->split_val);
          if(thain==1) {
             printf("( ");
             for(j=0;j<num_classes;j++) {
                printf("%s:%d ",classes_list[j],tree_root->children[i]->class_dist[j]);
             }
             printf(")");
          }
	}
	if(tree_root->children[i]->leaf==0) {
	  printf("\n");
	}
	print_level_cont(tree_root->children[i],1);
      }
      else { //NOMINAL
	printf("%s = %s ",attr_name_list[tree_root->attribute],attr_list[tree_root->attribute][i]);
        if(thain==1) {
           printf("( ");
           for(j=0;j<num_classes;j++) {
              printf("%s:%d ",classes_list[j],tree_root->children[i]->class_dist[j]);
           }
           printf(")");
        }

	if(tree_root->children[i]->leaf==0) {
	  //printf("\n");
	}
	print_level_cont(tree_root->children[i],1);
      }
    }
  //}
}

void print_splits()
{
  int i,j;
  for(i=0;i<num_attrs;i++) {
    if(attr_types[i] == CONTINUOUS) {
      printf("splits for attr %d:\n\n",i);
      for(j=0;j<num_splits[i];j++) {
	printf("%f\n",splits_list[i][j]);
      }
      printf("======================\n\n");
    }
  }
}

void print_level(struct tree_node * t, int level)
{
  int i,j;
  if(t->leaf==1) {
    printf("( ");
    for(i=0;i<num_classes;i++) {
      printf("%s:%d ",classes_list[i],t->class_dist[i]);
    }
    printf(")\n");
  }
  else {
    for(i=0;i<t->num_children;i++) {
      for(j=0;j<level;j++) {
	printf("| ");
      }
      printf("%s = %s ",attr_name_list[t->attribute],attr_list[t->attribute][i]);

      if(t->children[i]->leaf==0) {
	printf("\n");
      }
      print_level(t->children[i] ,level+1);
    }
  }
}

void print_level_cont(struct tree_node * t, int level)
{
  int i,j;
  if(t->leaf==1) {
    printf("( ");
    for(i=0;i<num_classes;i++) {
      printf("%s:%d ",classes_list[i],t->class_dist[i]);
    }
    printf(")\n");
  }
  else {
    for(i=0;i<t->num_children;i++) {
      for(j=0;j<level;j++) {
	printf("| ");
      }
      if(attr_types[t->attribute]==CONTINUOUS) {
	if(i==0) {
	  printf("%s <= %f ",attr_name_list[t->attribute],t->split_val);
	}
	else {
	  printf("%s > %f ",attr_name_list[t->attribute],t->split_val);
	}

	if(t->children[i]->leaf==0) {
          if(thain==1) {
             printf("( ");
             for(j=0;j<num_classes;j++) {
                printf("%s:%d ",classes_list[j],t->children[i]->class_dist[j]);
             }
             printf(")");
          }
          printf("\n");
        }
	print_level_cont(t->children[i] ,level+1);
      }
      else {
	printf("%s = %s ",attr_name_list[t->attribute],attr_list[t->attribute][i]);

	if(t->children[i]->leaf==0) {
	  //printf("\n");
          if(thain==1) {
             printf("( ");
             for(j=0;j<num_classes;j++) {
                printf("%s:%d ",classes_list[j],t->children[i]->class_dist[j]);
             }
             printf(")");
          }
          printf("\n");
        }
	print_level_cont(t->children[i] ,level+1);
      }
    }
  }
}

extern char *optarg;
extern int optind;
char options[]="hf:uv::m:c:PUX:M:LS";

/* Long options. This gives us enough room to add as many new features
   as we want without name collisions. */
enum {no_option, option_version,option_help,option_infgain,option_hellinger,option_hellinger_inv,option_chord,option_chisquaremetric,option_chisquaredistance,option_speciesprofile,option_euclidean,option_gini,option_dkm,option_gini_skewfree,option_thain,option_hellinger_weightoutside,option_hellinger_weightinside,option_hellinger_weightboth,option_dkm_avg,option_dkm_1againstavg,option_dkm_1againstmax,option_hellinger_1againstmax
};
struct option longopts[]=
{
  {"version",           no_argument,       0, option_version},
  {"help",              no_argument,       0, 'h'},

  /* Initial C4.5 options, extended to long options */
  {"filestem",          required_argument, 0, 'f'},
  {"test",              no_argument,       0, 'u'},
  {"verbosity",         optional_argument, 0, 'v'},
  {"min-objs",          required_argument, 0, 'm'},
  {"confidence",        required_argument, 0, 'c'},
  {"use-unpruned-tree",      no_argument,       0, 'U'},
  {"predict",                no_argument,       0, 'P'},
  {"leafs",                  no_argument,       0, 'L'},
  {"seed", required_argument,                  0, 'X'},
  {"depth", required_argument,                 0, 'M'},
  {"stats", no_argument,                 0, 'S'},
  {"pure",optional_argument,0,'Z'},  /*build pure trees*/
  /* splitting criteria */
  {"infgain",             no_argument,       0, option_infgain},
  {"gini",                no_argument,       0, option_gini},
  {"dkm",                 no_argument,       0, option_dkm},
  {"dkm-avg",             no_argument,       0, option_dkm_avg},
  {"dkm-1againstavg",     no_argument,       0, option_dkm_1againstavg},
  {"dkm-1againstmax",     no_argument,       0, option_dkm_1againstmax},
  {"gini-skewfree",       no_argument,       0, option_gini_skewfree},
  {"hellinger",           no_argument,       0, option_hellinger},
  {"hellinger-1againstmax", no_argument,     0, option_hellinger_1againstmax},
  {"hellinger-weightout", no_argument,       0, option_hellinger_weightoutside},
  {"hellinger-weightin",  no_argument,       0, option_hellinger_weightinside},
  {"hellinger-weightboth",no_argument,       0, option_hellinger_weightboth},
  {"hellinger-inv",       no_argument,       0, option_hellinger_inv},
  {"chord",               no_argument,       0, option_chord},
  {"chisquare-metric",    no_argument,       0, option_chisquaremetric},
  {"chisquare-distance",  no_argument,       0, option_chisquaredistance},
  {"species-profile",     no_argument,       0, option_speciesprofile},
  {"euclidean",           no_argument,       0, option_euclidean},
  {"thain",               no_argument,       0, option_thain},
  /* Last element must be all zeroes */
  {0,0,0,0}
};

char usage[]="\n"
"Usage: %s -f filestem\n"
"  Original C4.5 Options\n"
"   -h | --help               Print this screen\n"
"   --version                 Print version information and exit\n"
"\n"
"   -f s | --filestem s       Use s as filestem\n"
"\n"
"   Original options\n"
"   ================\n"
"   -b   | --batch            Use batch mode\n"
"   -u   | --test             Test trees against test set filestem.test\n"
"   -v n | --verbosity n      Use verbosity level n\n"
"\n"
"   File IO Options\n"
"   ===============\n"
"   -P | --predict            	Output predictions using best tree to \n"
"                             	filestem.pred\n"
"   -L | --leafs                Output leafs training and testing examples\n"
"                               fall under\n"
"   -S | --stats                Output tree height, number of nodes and leafs\n"
"   --thain                     Output distribution of each node\n"
"\n"
"\n"
"   Tree Growing \n"
"   ====================\n"
"   -M n | --depth n	     	Limit tree to a depth of n\n"
"   --collapse		     	Collapse trees (default)\n"
"      --no-collapse        \n"
"\n"
"  --pure                       Build pure trees\n"
"\n"
"   Pruning Flags\n"
"   =============\n"
"   -m n | --min-objs n          In all tests, at least two branches must\n"
"                                contain a minimum of n objects to split.\n"
"   -c n | --confidence n        Use n%% as confidence level\n"
"\n"
"   Splitting Criteria\n"
"   ==================\n"
"   --infgain               use Information Gain (default)\n"
"   --hellinger             use Hellinger Distance Metric\n"
"   --hellinger-inv         use Inverted Hellinger Distance Metric\n"
"   --chord                 use Chord Distance\n"
"   --chisquare-metric      use Chi-Square Metric\n"
"   --chisquare-distance    use Chi-Square Distance\n"
"   --species-profile       use Species Profile\n"
"   --euclidean             use Euclidean Distance\n"
"\n"
"\n"
;

int _process_options(int argc, char **argv)
{
  int o;
  long seed=0;
  int option_index;
  
  //PrintHeader("decision tree generator");
  
  /*  Process options  */
  while ( (o = getopt_long(argc, argv, options,longopts,&option_index)) != EOF ) {
    switch (o) {
    case 'L':
      OUTPUT_LEAFS = 1;
      printf("\tOutput Test set Leafs\n");
      break;
    case 'f':   
      filestem = strdup(optarg);
      printf("\tFile stem <%s>\n", filestem);
      break;
    case 'u':   
      evaluate_test_set = 1;
      LOAD_TEST_SET=1;
      printf("\tTrees evaluated on unseen cases\n");
      break;
    case 'v':   
      if ( optarg ) VERBOSITY = atoi(optarg);
      else          VERBOSITY = 1;
      printf("\tVerbosity level %d\n", VERBOSITY);
      break;
    case 'm':   
      MINOBJS = atoi(optarg);
      printf("\tSensible test requires 2 branches with >=%d cases\n",MINOBJS);
      if ( MINOBJS < 1 || MINOBJS > 1000000 )
         die("\t** illegal value **\n");
      break;
    case 'c':   
      confidence_factor = atof(optarg);
      printf("\tPruning confidence level %g%%\n", confidence_factor);
      if ( confidence_factor < Epsilon || confidence_factor > 100 )
         die("\t** illegal value **\n");
      confidence_factor /= 100;
      break;
    case 'U':
      use_pruning=0;
      printf("\tUsing unpruned tree for evaluation\n");
      break;
    case 'P':
      LOAD_TEST_SET=1;
      do_predictions=1;
      printf("\tOutput predictions\n");
      break;
    case 'X':
      seed=atol(optarg);
      printf("\tUsing %ld as the seed\n", seed);
      break;
    case 'M':
      max_tree_depth=atoi(optarg);
      printf("\tUsing %d as max tree depth\n", max_tree_depth);
      break;
    case 'Z':
      collapse_tree=0;
      MINOBJS = 1;
      PURE_TREES = 1;
      printf("\tPure Trees\n");
      break;
    case 'S':
      tree_stats=1;
      break;
    case option_infgain:
      gini_dist=0;
      dkm_dist=0;
      dkm_avg_dist=0;
      dkm_1againstavg_dist=0;
      dkm_1againstmax_dist=0;
      gini_skewfree_dist=0;
      purity=1;
      euclidean_dist=0;
      chord_dist=0;
      chisquaremetric_dist=0;
      chisquaredistance_dist=0;
      speciesprofile_dist=0;
      hellinger_dist=0;
      hellinger_weightboth_dist=0;
      hellinger_weightinside_dist=0;
      hellinger_weightoutside_dist=0;
      hellinger_inv_dist=0;
      purity=1;
      break;
    case option_gini:
      infgain_dist=0;
      gini_dist=1;
      dkm_dist=0;
      dkm_avg_dist=0;
      dkm_1againstavg_dist=0;
      dkm_1againstmax_dist=0;
      gini_skewfree_dist=0;
      purity=1;
      euclidean_dist=0;
      chord_dist=0;
      chisquaremetric_dist=0;
      chisquaredistance_dist=0;
      speciesprofile_dist=0;
      hellinger_dist=0;
      hellinger_weightboth_dist=0;
      hellinger_weightinside_dist=0;
      hellinger_weightoutside_dist=0;
      hellinger_inv_dist=0;
      purity=1;
      break;
    case option_dkm:
      infgain_dist=0;
      gini_dist=0;
      dkm_dist=1;
      dkm_avg_dist=0;
      dkm_1againstavg_dist=0;
      dkm_1againstmax_dist=0;
      gini_skewfree_dist=0;
      purity=1;
      euclidean_dist=0;
      chord_dist=0;
      chisquaremetric_dist=0;
      chisquaredistance_dist=0;
      speciesprofile_dist=0;
      hellinger_dist=0;
      hellinger_weightboth_dist=0;
      hellinger_weightinside_dist=0;
      hellinger_weightoutside_dist=0;
      hellinger_inv_dist=0;
      purity=1;
      break;
    case option_dkm_avg:
      infgain_dist=0;
      gini_dist=0;
      dkm_dist=0;
      dkm_avg_dist=1;
      dkm_1againstavg_dist=0;
      dkm_1againstmax_dist=0;
      gini_skewfree_dist=0;
      purity=1;
      euclidean_dist=0;
      chord_dist=0;
      chisquaremetric_dist=0;
      chisquaredistance_dist=0;
      speciesprofile_dist=0;
      hellinger_dist=0;
      hellinger_weightboth_dist=0;
      hellinger_weightinside_dist=0;
      hellinger_weightoutside_dist=0;
      hellinger_inv_dist=0;
      purity=1;
      break;
    case option_dkm_1againstavg:
      infgain_dist=0;
      gini_dist=0;
      dkm_dist=0;
      dkm_avg_dist=0;
      dkm_1againstavg_dist=1;
      dkm_1againstmax_dist=0;
      gini_skewfree_dist=0;
      purity=1;
      euclidean_dist=0;
      chord_dist=0;
      chisquaremetric_dist=0;
      chisquaredistance_dist=0;
      speciesprofile_dist=0;
      hellinger_dist=0;
      hellinger_weightboth_dist=0;
      hellinger_weightinside_dist=0;
      hellinger_weightoutside_dist=0;
      hellinger_inv_dist=0;
      purity=1;
      break;
    case option_dkm_1againstmax:
      infgain_dist=0;
      gini_dist=0;
      dkm_dist=0;
      dkm_avg_dist=0;
      dkm_1againstavg_dist=0;
      dkm_1againstmax_dist=1;
      gini_skewfree_dist=0;
      purity=1;
      euclidean_dist=0;
      chord_dist=0;
      chisquaremetric_dist=0;
      chisquaredistance_dist=0;
      speciesprofile_dist=0;
      hellinger_dist=0;
      hellinger_weightboth_dist=0;
      hellinger_weightinside_dist=0;
      hellinger_weightoutside_dist=0;
      hellinger_inv_dist=0;
      purity=1;
      break;
    case option_gini_skewfree:
      infgain_dist=0;
      gini_dist=0;
      dkm_dist=0;
      dkm_avg_dist=0;
      dkm_1againstavg_dist=0;
      dkm_1againstmax_dist=0;
      gini_skewfree_dist=1;
      purity=1;
      euclidean_dist=0;
      chord_dist=0;
      chisquaremetric_dist=0;
      chisquaredistance_dist=0;
      speciesprofile_dist=0;
      hellinger_dist=0;
      hellinger_weightboth_dist=0;
      hellinger_weightinside_dist=0;
      hellinger_weightoutside_dist=0;
      hellinger_inv_dist=0;
      purity=1;
      break;
    case option_hellinger:
      infgain_dist=0;
      gini_dist=0;
      dkm_dist=0;
      dkm_avg_dist=0;
      dkm_1againstavg_dist=0;
      dkm_1againstmax_dist=0;
      gini_skewfree_dist=0;
      euclidean_dist=0;
      chord_dist=0;
      chisquaremetric_dist=0;
      chisquaredistance_dist=0;
      speciesprofile_dist=0;
      hellinger_dist=1;
      hellinger_weightboth_dist=0;
      hellinger_weightinside_dist=0;
      hellinger_weightoutside_dist=0;
      hellinger_inv_dist=0;
      purity=0;
      break;
    case option_hellinger_1againstmax:
      hellinger_1againstmax_dist=1;
      infgain_dist=0;
      gini_dist=0;
      dkm_dist=0;
      dkm_avg_dist=0;
      dkm_1againstavg_dist=0;
      dkm_1againstmax_dist=0;
      gini_skewfree_dist=0;
      euclidean_dist=0;
      chord_dist=0;
      chisquaremetric_dist=0;
      chisquaredistance_dist=0;
      speciesprofile_dist=0;
      hellinger_dist=0;
      hellinger_weightboth_dist=0;
      hellinger_weightinside_dist=0;
      hellinger_weightoutside_dist=0;
      hellinger_inv_dist=0;
      purity=0;
    case option_hellinger_weightboth:
      infgain_dist=0;
      gini_dist=0;
      dkm_dist=0;
      dkm_avg_dist=0;
      dkm_1againstavg_dist=0;
      dkm_1againstmax_dist=0;
      gini_skewfree_dist=0;
      euclidean_dist=0;
      chord_dist=0;
      chisquaremetric_dist=0;
      chisquaredistance_dist=0;
      speciesprofile_dist=0;
      hellinger_dist=0;
      hellinger_weightboth_dist=1;
      hellinger_weightinside_dist=0;
      hellinger_weightoutside_dist=0;
      hellinger_inv_dist=0;
      purity=0;
      break;
    case option_hellinger_weightinside:
      infgain_dist=0;
      gini_dist=0;
      dkm_dist=0;
      dkm_avg_dist=0;
      dkm_1againstavg_dist=0;
      dkm_1againstmax_dist=0;
      gini_skewfree_dist=0;
      euclidean_dist=0;
      chord_dist=0;
      chisquaremetric_dist=0;
      chisquaredistance_dist=0;
      speciesprofile_dist=0;
      hellinger_dist=0;
      hellinger_weightboth_dist=0;
      hellinger_weightinside_dist=1;
      hellinger_weightoutside_dist=0;
      hellinger_inv_dist=0;
      purity=0;
      break;
    case option_hellinger_weightoutside:
      infgain_dist=0;
      gini_dist=0;
      dkm_dist=0;
      dkm_avg_dist=0;
      dkm_1againstavg_dist=0;
      dkm_1againstmax_dist=0;
      gini_skewfree_dist=0;
      euclidean_dist=0;
      chord_dist=0;
      chisquaremetric_dist=0;
      chisquaredistance_dist=0;
      speciesprofile_dist=0;
      hellinger_dist=0;
      hellinger_weightboth_dist=0;
      hellinger_weightinside_dist=0;
      hellinger_weightoutside_dist=1;
      hellinger_inv_dist=0;
      purity=0;
      break;
    case option_hellinger_inv:
      infgain_dist=0;
      gini_dist=0;
      dkm_dist=0;
      dkm_avg_dist=0;
      dkm_1againstavg_dist=0;
      dkm_1againstmax_dist=0;
      gini_skewfree_dist=0;
      euclidean_dist=0;
      chord_dist=0;
      chisquaremetric_dist=0;
      chisquaredistance_dist=0;
      speciesprofile_dist=0;
      hellinger_dist=0;
      hellinger_weightboth_dist=0;
      hellinger_weightinside_dist=0;
      hellinger_weightoutside_dist=0;
      hellinger_inv_dist=1;
      purity=0;
      break;
    case option_chord:
      infgain_dist=0;
      gini_dist=0;
      dkm_dist=0;
      dkm_avg_dist=0;
      dkm_1againstavg_dist=0;
      dkm_1againstmax_dist=0;
      gini_skewfree_dist=0;
      euclidean_dist=0;
      chord_dist=1;
      chisquaremetric_dist=0;
      chisquaredistance_dist=0;
      speciesprofile_dist=0;
      hellinger_dist=0;
      hellinger_weightboth_dist=0;
      hellinger_weightinside_dist=0;
      hellinger_weightoutside_dist=0;
      hellinger_inv_dist=0;
      purity=0;
      break;
    case option_chisquaremetric:
      infgain_dist=0;
      gini_dist=0;
      dkm_dist=0;
      dkm_avg_dist=0;
      dkm_1againstavg_dist=0;
      dkm_1againstmax_dist=0;
      gini_skewfree_dist=0;
      euclidean_dist=0;
      chord_dist=0;
      chisquaremetric_dist=1;
      chisquaredistance_dist=0;
      speciesprofile_dist=0;
      hellinger_dist=0;
      hellinger_weightboth_dist=0;
      hellinger_weightinside_dist=0;
      hellinger_weightoutside_dist=0;
      hellinger_inv_dist=0;
      purity=0;
      break;
    case option_chisquaredistance:
      infgain_dist=0;
      gini_dist=0;
      dkm_dist=0;
      dkm_avg_dist=0;
      dkm_1againstavg_dist=0;
      dkm_1againstmax_dist=0;
      gini_skewfree_dist=0;
      euclidean_dist=0;
      chord_dist=0;
      chisquaremetric_dist=0;
      chisquaredistance_dist=1;
      speciesprofile_dist=0;
      hellinger_dist=0;
      hellinger_weightboth_dist=0;
      hellinger_weightinside_dist=0;
      hellinger_weightoutside_dist=0;
      hellinger_inv_dist=0;
      purity=0;
      break;
    case option_speciesprofile:
      infgain_dist=0;
      gini_dist=0;
      dkm_dist=0;
      dkm_avg_dist=0;
      dkm_1againstavg_dist=0;
      dkm_1againstmax_dist=0;
      gini_skewfree_dist=0;
      euclidean_dist=0;
      chord_dist=0;
      chisquaremetric_dist=0;
      chisquaredistance_dist=0;
      speciesprofile_dist=1;
      hellinger_dist=0;
      hellinger_weightboth_dist=0;
      hellinger_weightinside_dist=0;
      hellinger_weightoutside_dist=0;
      hellinger_inv_dist=0;
      purity=0;
      break;
    case option_euclidean:
      infgain_dist=0;
      gini_dist=0;
      dkm_dist=0;
      dkm_avg_dist=0;
      dkm_1againstavg_dist=0;
      dkm_1againstmax_dist=0;
      gini_skewfree_dist=0;
      euclidean_dist=1;
      chord_dist=0;
      chisquaremetric_dist=0;
      chisquaredistance_dist=0;
      speciesprofile_dist=0;
      hellinger_dist=0;
      hellinger_weightboth_dist=0;
      hellinger_weightinside_dist=0;
      hellinger_weightoutside_dist=0;
      hellinger_inv_dist=0;
      purity=0;
      break;
    case option_thain:
      thain=1;
      break;
    case '?':   printf("unrecognised option: %c,\n", (char) o);
    case option_version:
      fprintf(stdout,"Version Unknown\n");
      //fprintf(stdout,"%s\n",version);
      exit(1);
    case 'h':case option_help:
      fprintf(stdout,usage);
      //fprintf(stdout,usage,progname);
      exit(1);
    }
  }

  /*
  if ( divide_and_conquer ) {
    LOAD_TEST_SET=True;
    use_pruning=0;
  }
  */

  /* 
     Initialize random number generator. If seed is set to 0 (default),
     use time otherwise use the seed.
  */
  if ( seed == 0 ) {
    seed = time(NULL)+getpid();
  } 
  srand48(seed);
  printf("\tRandom seed is %ld\n", seed);
  
  return 0;
}




/*************************************************************************/
/*                                                                       */
/*  Compute the total information in V[ MinVal..MaxVal ]                 */
/*                                                                       */
/*************************************************************************/
float TotalInfo(float V[], int min_value, int max_value)
{
  int v;
  float Sum=0.0;
  float N, TotalItems=0.0;

  for (v=min_value; v < max_value; v++) {
    N = V[v];
    
    Sum+= N * Log(N);
    printf("%.12f\n",Sum);
    TotalItems += N;
  }


  printf("%.12f %.12f %.12f\n",TotalItems, Log(TotalItems), TotalItems * Log(TotalItems));
  N = TotalItems * Log(TotalItems) - Sum;
  printf("%.12f\n",N);
  return N;

}
