/*************************************************************************/
/*                                                              	 */
/*  Determine the class of a case description from a decision tree	 */
/*  --------------------------------------------------------------	 */
/*                                                              	 */
/*************************************************************************/


#include "defns.h"
#include "types.h"
#include "extern.h"

float	*ClassSum=NULL;		/* ClassSum[c] = total weight of class c */
extern Tree conquer_node;


/*************************************************************************/
/*                                                              	 */
/*  Categorize a case description using the given decision tree		 */
/*                                                              	 */
/*************************************************************************/
int Category(Description CaseDesc, TREE *DecisionTree) 
{ 
  int c, best_class;
  
  if ( ! ClassSum ) { 
    ClassSum=MALLOC(number_of_classes * sizeof(float));
  }

  for (c=0; c < number_of_classes; c++) 
    ClassSum[c] = 0;

  
  Classify(CaseDesc, DecisionTree, 1.0);

  best_class = 0;
  for (c=0; c < number_of_classes; c++) {
    Verbosity(5) printf("class %s weight %.2f\n", class_name[c], ClassSum[c]);
    if ( ClassSum[c] > ClassSum[best_class] ) best_class = c;
  }
  
  return best_class;
}


/*************************************************************************/
/*                                                              	 */
/*  Categorize a case description using the given decision tree		 */
/*                                                              	 */
/*************************************************************************/
int GetLeafID(Description CaseDesc, TREE *DecisionTree) 
{ 
  int c, best_class;
  
  if ( ! ClassSum ) { 
    ClassSum=MALLOC(number_of_classes * sizeof(float));
  }

  for (c=0; c < number_of_classes; c++) 
    ClassSum[c] = 0;

  
  return LeafID(CaseDesc, DecisionTree, 1.0);

  /*best_class = 0;
  for (c=0; c < number_of_classes; c++) {
    Verbosity(5) printf("class %s weight %.2f\n", class_name[c], ClassSum[c]);
    if ( ClassSum[c] > ClassSum[best_class] ) best_class = c;
    }*/
  
  //return best_class;
}


/*************************************************************************/
/*                                                              	 */
/*  Classify a case description using the given subtree by adjusting	 */
/*  the value ClassSum for each class					 */
/*                                                              	 */
/*************************************************************************/
void Classify(Description CaseDesc, TREE *T, float Weight)
{
  int v, dv;
  int a, c, d, newfile=0;
  float Cv, total=0.0;
  fpos_t pos;

  /*dcieslak add*/
  FILE * fp;
  char buf[1024], tmp[1024];

  switch ( T->node_type ) {
  case LEAF_NODE:  /* leaf */
    if ( T->Items > 0 ) {
      /*  Update from ALL classes  */
      for (c=0; c < number_of_classes; c++) {
	if ( T->class_distribution[c] ) 
	  ClassSum[c] += Weight * T->class_distribution[c] / T->Items;
      }
    } else {
      ClassSum[T->leaf] += Weight;
    }
    conquer_node=T;

    sprintf(buf,"%s.laplace",filestem);
    fp=fopen(buf,"r");
    if(fscanf(fp,"%s" ,tmp) == EOF)
      {
	newfile=1;
      }
    else
      {
	fclose(fp);
      }
    if(newfile)
      {
	fp=fopen(buf,"w");
	for (d=0; d < number_of_classes; d++) {
	  fprintf(fp,"%s",class_name[d]);
	  fprintf(fp,"%c",(d==number_of_classes-1)?'\n':',');
	}
      }
    else
      {
	fp=fopen(buf,"a");
      }
    for (c=0; c < number_of_classes; c++) {
      total+=T->class_distribution[c];
    }
    for (c=0; c < number_of_classes; c++) {
      fprintf(fp,"%.6f",(double)(T->class_distribution[c]+1.0)/(total+1.0*number_of_classes));
      //fprintf(fp,"%f",T->class_distribution[c]);
      fprintf(fp,"%c",(c==number_of_classes-1)?'\n':',');
    }
    fclose(fp);

    if(OUTPUT_LEAFS)
      {
	FILE * fp;
	char buf[1024];
	sprintf(buf,"%s.leafs",filestem);
	fp=fopen(buf,"a");
	fprintf(fp,"%d\n",T->nodeid);
	//fprintf(stderr,"%s\n",filestem);
	//fprintf(stderr,"%d\n",T->nodeid);
	fclose(fp);
      }
    return;

  case DISCRETE_NODE:  /* test of discrete attribute */
    a = T->tested_attribute;
    v = DVal(CaseDesc, a);
    if ( v && v <= T->num_branches )	/*  Make sure not new discrete value  */
      Classify(CaseDesc, T->branch[v], Weight);
    else {
      for (v=1; v <= T->num_branches; v++) {
	Classify(CaseDesc, T->branch[v], 
		 (Weight * T->branch[v]->Items) / T->Items);
      }
    }
      return;
  case CONTINUOUS_NODE:  /* test of continuous attribute */
    a = T->tested_attribute;
    Cv = CVal(CaseDesc, a);

    if ( Cv == Unknown ) {
      for (v=1; v <= 2; v++) {
	Classify(CaseDesc, T->branch[v], 
		 (Weight * T->branch[v]->Items) / T->Items);
      }
    } else {
      v = ( Cv <= T->Cut ? 1 : 2 );
      Classify(CaseDesc, T->branch[v], Weight);
    }
    
    return;
  case SUBSET_NODE:  /* subset test on discrete attribute  */
    a = T->tested_attribute;
    dv = DVal(CaseDesc, a);
    
    if ( dv ) {
      for (v=1; v <= T->num_branches; v++) {
	if ( In(dv, T->Subset[v]) ) {
	  Classify(CaseDesc, T->branch[v], Weight);
	  
	  return;
	}
      }
    }

    /*  Value unknown or not found in any of the subsets  */
    for (v=1; v <= T->num_branches; v++) {
      Classify(CaseDesc, T->branch[v], 
	       (Weight * T->branch[v]->Items) / T->Items);
    }
    
    return;
  } 
}

/*************************************************************************/
/*                                                              	 */
/*  	Find which leaf a given example belongs to                       */
/*                                                              	 */
/*************************************************************************/
int LeafID(Description CaseDesc, TREE *T, float Weight)
{
  int v, dv;
  int a, c;
  float Cv;


  
  switch ( T->node_type ) {
  case LEAF_NODE:  /* leaf */
    if ( T->Items > 0 ) {
      /*  Update from ALL classes  */
      for (c=0; c < number_of_classes; c++) {
	if ( T->class_distribution[c] ) 
	  ClassSum[c] += Weight * T->class_distribution[c] / T->Items;
      }
    } else {
      ClassSum[T->leaf] += Weight;
    }
    conquer_node=T;
    /*if (OUTPUT_LEAFS) {
      fprintf(stderr,"%d\n",nodeid);
      }*/
    
    return T->nodeid;

  case DISCRETE_NODE:  /* test of discrete attribute */
    a = T->tested_attribute;
    v = DVal(CaseDesc, a);
    if ( v && v <= T->num_branches )	/*  Make sure not new discrete value  */
      return LeafID(CaseDesc, T->branch[v], Weight);
    else {
      for (v=1; v <= T->num_branches; v++) {
	return LeafID(CaseDesc, T->branch[v], 
		 (Weight * T->branch[v]->Items) / T->Items);
      }
    }
    //return;
  case CONTINUOUS_NODE:  /* test of continuous attribute */
    a = T->tested_attribute;
    Cv = CVal(CaseDesc, a);

    if ( Cv == Unknown ) {
      for (v=1; v <= 2; v++) {
	return LeafID(CaseDesc, T->branch[v], 
		 (Weight * T->branch[v]->Items) / T->Items);
      }
    } else {
      v = ( Cv <= T->Cut ? 1 : 2 );
      return LeafID(CaseDesc, T->branch[v], Weight);
    }
    
    //return;
  case SUBSET_NODE:  /* subset test on discrete attribute  */
    a = T->tested_attribute;
    dv = DVal(CaseDesc, a);
    
    if ( dv ) {
      for (v=1; v <= T->num_branches; v++) {
	if ( In(dv, T->Subset[v]) ) {
	  return LeafID(CaseDesc, T->branch[v], Weight);
	  
	  //return;
	}
      }
    }

    /*  Value unknown or not found in any of the subsets  */
    for (v=1; v <= T->num_branches; v++) {
      return LeafID(CaseDesc, T->branch[v], 
	       (Weight * T->branch[v]->Items) / T->Items);
    }
    
    //return;
  } 
}
