/*
  create_trees.c

  Creates decision trees from the data. A separate file (windowing)
  handles the use of windowing. Here, we start off the creation of
  one or more trees.
*/
#include "defns.h"
#include "types.h"
#include "extern.h"


extern int best_tree;
extern Boolean LOAD_UNPRUNED_TREES, LOAD_PRUNED_TREES;
extern Boolean PRUNING_REQUESTED;
extern int nodeid;
extern Boolean peel_off;
void create_trees()
{
  int i,err;
  int height, nodes, leafs;
  FILE * fp;
  char filename[MAXBUF];

  /* Set up tree building variables */
  InitialiseTreeData();

  /* No where else to put it, initialize oob estimates if necessary */
  if ( use_oob ) init_oob();

  /* initialize boosting crap */
  if ( use_boosting ) init_boosting();

  /* Allocate pruned and unpruned storage */
  if ( ! LOAD_UNPRUNED_TREES && ! LOAD_PRUNED_TREES ) 
    unpruned_trees=CALLOC(number_of_trees,sizeof(TREE *));
  if ( PRUNING_REQUESTED && ! LOAD_PRUNED_TREES )
    pruned_trees=CALLOC(number_of_trees,sizeof(TREE *));

  info("Building tree\n");
  /* Go through and build (and prune) each tree */
  for (i=0; i < number_of_trees; i++) {
    info("%d ",i+1);

    /* Intializations each time */
    TRIALS=1;
    AllKnown=True;
    InitialiseWeights();
    nodeid=0;                   /* A counter of nodes in the tree */

    /* Initialize Items with training set */
    if ( use_bagging )
      load_bagged_data();
    else if ( use_boosting )
      load_boosted_data();
    else
      load_training_data();

    /* If we use random subspaces, pick them here */
    if ( use_random_subspaces )
      initialize_random_subspaces();
    if ( peel_off ) initialize_peel_off();
    /* Possibly create the tree */
    if ( ! LOAD_UNPRUNED_TREES && ! LOAD_PRUNED_TREES) {
      /*
	Create a tree with items from 0 to number_of_examples total,
	starting at depth 0.
      */
      if ( use_windowing )
	best_tree=BestTree();
      else
	unpruned_trees[i]=FormTree(0,number_of_examples,0);

      /* Special case print the tree out */
      if ( number_of_trees == 1 ) print_tree(unpruned_trees[0]);

      /* Optionally include oob error estimates */
      if ( use_oob ) 
	calculate_oob(unpruned_trees[i]);

    }
   
    /* Now prune, if need be */
    if ( PRUNING_REQUESTED && ! LOAD_PRUNED_TREES )
      pruned_trees[i]=prune_tree(unpruned_trees[i]);

    if ( use_boosting )
    {
      err=BoostingUpdateWeights(i);
      if (err==-1) /* Error was too great, kill the classifier */
      {
        i--;
        continue;
      }
      if (err==-2) /* Error was zero... stop building now */
        number_of_trees=i+1;
    }

  }

  if(tree_stats==1) {
     height = find_tree_stats(unpruned_trees[0], &nodes, &leafs);
     bzero(filename,MAXBUF);
     sprintf(filename,"%s.tree_info", filestem);
     fp=fopen(filename,"w");
     fprintf(fp,"height: %d\n",height);
     fprintf(fp,"#nodes: %d\n",nodes);
     fprintf(fp,"#leafs: %d\n",leafs);
     fclose(fp);
   }


  info("\n");
  if ( ! use_windowing ) 
    best_tree=0;
}



int find_tree_stats(TREE *tree, int * nodes, int * leafs) {
  int i, height, max;
  (*nodes)++;
  printf("top\n");
  printf("node type %d\n",tree->node_type);
  if(tree->node_type==LEAF_NODE) {
     (*leafs)++;
     return 0;
  } 
  else {
     if(tree->node_type==CONTINUOUS_NODE) {
        printf("left\n");
        height = find_tree_stats(tree->branch[1], nodes, leafs);
        printf("right\n");
        max = find_tree_stats(tree->branch[2], nodes, leafs);
        if(max > height) {
           height = max;
        }
     }
     else { //NOMINAL 
        max = 0;
        for(i=1;i<=tree->num_branches;i++) {
           height = find_tree_stats(tree->branch[i], nodes, leafs);
           if(height > max) {
              max = height;
           }
        }
        height = max;
     }
     printf("returning %d\n",height+1);
     return (height + 1);
  }
}














