#include "distance.h"

void create_tree() {
  int i, *used_train, *used_attr, nodeid=1, height=0, nodes=0, leafs=0;
  FILE * fp;
  char filename[MAXLINE];

  used_train=malloc(num_train*sizeof(int));
  for(i=0;i<num_train;i++) {
    used_train[i]=0;
  }

  used_attr=malloc(num_attrs*sizeof(int));
  for(i=0;i<num_attrs;i++) {
    used_attr[i]=0;
    if(attr_types[i]==CONTINUOUS) {
      nominal_only=0;
    }
  }

  tree_root = malloc(sizeof(struct tree_node));
  /*if(nominal_only==1) {
    printf("using ID3\n");
    make_tree_ID3(&tree_root, used_train, used_attr, &nodeid);
  }
  else {*/
    printf("using C4.5\n");
    make_tree_C45(&tree_root, used_train, used_attr, &nodeid, 0);
  //}

  if(tree_stats==1) {
    height = find_tree_stats(tree_root, &nodes, &leafs);
    bzero(filename,MAXLINE);
    sprintf(filename,"%s.tree_info", filestem);
    fp=fopen(filename,"w");
    fprintf(fp,"height: %d\n",height);
    fprintf(fp,"#nodes: %d\n",nodes);
    fprintf(fp,"#leafs: %d\n",leafs);
    fclose(fp);
  }

  

  free(used_attr);
  free(used_train);
}

int find_tree_stats(struct tree_node * tree, int * nodes, int * leafs) {
  int i, height, max;
  (*nodes)++;
  if(tree->leaf==1) {
     (*leafs)++;
     return 0;
  } 
  else {
     if(attr_types[tree->attribute]==CONTINUOUS) {
        height = find_tree_stats(tree->children[0], nodes, leafs);
        max = find_tree_stats(tree->children[1], nodes, leafs);
        if(max > height) {
           height = max;
        }
     }
     else { //NOMINAL 
        max = 0;
        for(i=0;i<num_attr_vals[tree->attribute];i++) {
           height = find_tree_stats(tree->children[i], nodes, leafs);
           if(height > max) {
              max = height;
           }
        }
        height = max;
     }
     return (height + 1);
  }
}

void make_tree_ID3(struct tree_node ** tree, int * used_examples, int * used_attr, int * nodeid) {
  int i,j,best,*new_used_examples,*new_used_attr,count=0;

  for(j=0;j<num_train;j++) {
    if(used_examples[j] == 0) {
      count++;
    }
  }

  best=best_ID3(used_examples,used_attr);
  if(best==-1 || count < MINOBJS) {
    int max;
    (*tree)->leaf=1;
    (*tree)->class_dist=malloc(sizeof(int)*num_classes);
    for(i=0;i<num_classes;i++) {
      (*tree)->class_dist[i]=0;
    }
    for(i=0;i<num_train;i++) {
      if(used_examples[i]==0) {
	(*tree)->class_dist[train_class_list[i]]++;
      }
    }
    (*tree)->nodeid=(*nodeid);
    (*nodeid)=(*nodeid)+1;
    max=(*tree)->class_dist[0];
    (*tree)->class=0;
    for(i=1;i<num_classes;i++) {
      if((*tree)->class_dist[i] > max) {
	max=(*tree)->class_dist[i];
	(*tree)->class=i;
      }
      //(*tree)->class_dist[i]=0;
    }
  }
  else {
    (*tree)->attribute=best;
    (*tree)->leaf=0;
    (*tree)->num_children=num_attr_vals[best];
    (*tree)->children=malloc(num_attr_vals[best]*sizeof(struct tree_node *));
    
    new_used_attr=malloc(num_attrs*sizeof(int));
    for(i=0;i<num_attrs;i++) {
      new_used_attr[i]=used_attr[i];
    }

    new_used_attr[best]=1;
    for(i=0;i<num_attr_vals[best];i++) {
      new_used_examples=malloc(num_train*sizeof(int));
      for(j=0;j<num_train;j++) {
	new_used_examples[j]=used_examples[j];
	if(train_attr_list[j][best] != i) {
	  new_used_examples[j]=1;
	}
      }
      (*tree)->children[i]=malloc(sizeof(struct tree_node));
      make_tree_ID3(&((*tree)->children[i]), new_used_examples, new_used_attr, nodeid);
    }
  }
}

//return the index of the best attribute based on chosen distance
//I rewrote this function on 031207 to use the libdistance.a funcs
int best_ID3(int * used_examples, int * used_attr)
{

  int i,j,k,best=-1, **N;
  float float_best=0.0, *distance;

  distance=malloc(num_attrs*sizeof(float));
  for(i=0;i<num_attrs;i++) {
    distance[i]=0.0;
  }

  for(i=0;i<num_attrs;i++) {
    if(used_attr[i]==0)	{
      N=malloc(num_classes*sizeof(int *));
      for(j=0;j<num_classes;j++) {
	N[j]=malloc(num_attr_vals[i]*sizeof(int));
	for(k=0;k<num_attr_vals[i];k++) {
	  N[j][k]=0;
	}
      }
      
      for(j=0;j<num_train;j++) {
	if(used_examples[j]==0) {
	  N[train_class_list[j]][train_attr_list[j][i]]++;
	}
      }
      distance[i]=calcdistance(N,num_classes,2,used_examples);
      for(j=0;j<num_classes;j++) {
	free(N[j]);
	    }
    }
  }
  /* find the best distance and return it */
  for(i=0;i<num_attrs;i++) {
    if(distance[i]>float_best) {
      best=i;
      float_best=distance[i];
    }
  }
  return best;
}

void make_tree_C45(struct tree_node ** tree, int * used_examples, int * used_attr, int * nodeid, int current_tree_depth) {
  int i,j,k,best,*new_used_examples,*new_used_attr,count=0,pure=2;
  float split_val=0.0,avg_distance=0.0,*inf;

  int * class_count;
  class_count=malloc(sizeof(int)*num_classes);
  inf        =malloc(sizeof(float)*num_attrs);
  best=best_C45(used_examples,used_attr,&split_val,inf,&avg_distance);
  //printf("best attribute: %d\n",best);
  for(j=0;j<num_classes;j++) {
    class_count[j]=0;
  }

  for(j=0;j<num_train;j++) {
    if(used_examples[j] == 0) {
      count++;
      class_count[train_class_list[j]]++;
    }
  }

  for(j=0;j<num_classes;j++) {
    if(class_count[j]>0) {
      pure--;
    }
  }

  free(class_count);
  if(best==-1 || count < 2*MINOBJS || pure > 0
     || ( max_tree_depth >= 0 && current_tree_depth >= max_tree_depth ) ) {
    int max,total=0;
    (*tree)->leaf=1;
    (*tree)->class_dist=malloc(sizeof(int)*num_classes);
    for(i=0;i<num_classes;i++) {
      (*tree)->class_dist[i]=0;
    }
    for(i=0;i<num_train;i++) {
      if(used_examples[i]==0) {
	(*tree)->class_dist[train_class_list[i]]++;
      }
    }
    (*tree)->nodeid=(*nodeid);
    (*nodeid)=(*nodeid)+1;
    max=(*tree)->class_dist[0];
    (*tree)->class=0;
    total+=max;
    for(i=1;i<num_classes;i++) {
      if((*tree)->class_dist[i] > max) {
	total+=(*tree)->class_dist[i];
	max=(*tree)->class_dist[i];
	(*tree)->class=i;
      }
    }
    (*tree)->errors=total-max;
  }
  else {
    (*tree)->attribute=best;
    if(attr_types[best]==CONTINUOUS) {
      (*tree)->leaf=0;
      (*tree)->split_val=split_val;
      (*tree)->num_children=2;
      (*tree)->children=malloc(2*sizeof(struct tree_node *));
      (*tree)->errors=0;
      (*tree)->class_dist=malloc(sizeof(int)*num_classes);
      (*tree)->nodeid=(*nodeid);
      (*nodeid)=(*nodeid)+1;

      if(thain==1) {
         (*tree)->class_dist=malloc(sizeof(int)*num_classes);
         for(i=0;i<num_classes;i++) {
            (*tree)->class_dist[i]=0;
         }
         for(i=0;i<num_train;i++) {
            if(used_examples[i]==0) {
               (*tree)->class_dist[train_class_list[i]]++;
            }
         }
      }


      for(k=0;k<num_classes;k++) {
	(*tree)->class_dist[k]=0;	      
      }

      for(i=0;i<2;i++) {
	new_used_examples=malloc(num_train*sizeof(int));
	for(j=0;j<num_train;j++) {
	  new_used_examples[j]=used_examples[j];
	  if(train_continuous_list[j][best] > split_val && i == 0) {
	    new_used_examples[j]=1;
	  }
	  else if(train_continuous_list[j][best] <= split_val && i == 1) {
	    new_used_examples[j]=1;
	  }
	}
	(*tree)->children[i]=malloc(sizeof(struct tree_node));

	make_tree_C45(&((*tree)->children[i]), new_used_examples, used_attr, nodeid, current_tree_depth+1);
	(*tree)->errors+=(*tree)->children[i]->errors;
	for(k=0;k<num_classes;k++) {
	  (*tree)->class_dist[k]+=(*tree)->children[i]->class_dist[k];
	}
	free(new_used_examples);
      }
      int num_cases=0, best_class_count=0;
      for(k=0;k<num_classes;k++) {
	if( (*tree)->class_dist[k] > best_class_count ) {
	  best_class_count = (*tree)->class_dist[k];
	  (*tree)->class=k;
	}
	num_cases+=(*tree)->class_dist[k];
      }
      if ( collapse_tree==1 && (*tree)->errors >= num_cases - best_class_count - Epsilon ) {
	(*tree)->leaf=1;
	(*tree)->num_children=0;
      }
    }
    else if (attr_types[best] == NOMINAL){
      (*tree)->leaf=0;
      (*tree)->num_children=num_attr_vals[best];
      (*tree)->children=malloc(num_attr_vals[best]*sizeof(struct tree_node *));
      (*tree)->errors=0;
      new_used_attr=malloc(num_attrs*sizeof(int));
      (*tree)->class_dist=malloc(sizeof(int)*num_classes);
      (*tree)->nodeid=(*nodeid);
      (*nodeid)=(*nodeid)+1;
      for(k=0;k<num_classes;k++) {
	(*tree)->class_dist[k]=0;	      
      }
      for(i=0;i<num_attrs;i++) {
	new_used_attr[i]=used_attr[i];
      }
      new_used_attr[best]=1;
	  
      for(i=0;i<num_attr_vals[best];i++) {
	new_used_examples=malloc(num_train*sizeof(int));
	for(j=0;j<num_train;j++) {
	  new_used_examples[j]=used_examples[j];
	  if(train_attr_list[j][best] != i) {
	    new_used_examples[j]=1;
	  }
	}
	(*tree)->children[i]=malloc(sizeof(struct tree_node));
	make_tree_C45(&((*tree)->children[i]), new_used_examples, new_used_attr, nodeid, current_tree_depth+1);
	(*tree)->errors+=(*tree)->children[i]->errors;
	for(k=0;k<num_classes;k++) {
	  (*tree)->class_dist[k]+=(*tree)->children[i]->class_dist[k];
	}
	free(new_used_examples);
      }
      free(new_used_attr);
      int num_cases=0, best_class_count=0;
      for(k=0;k<num_classes;k++) {
	if( (*tree)->class_dist[k] > best_class_count ) {
	  best_class_count = (*tree)->class_dist[k];
	  (*tree)->class=k;
	}
	      num_cases+=(*tree)->class_dist[k];
      }
      if ( collapse_tree==1 && (*tree)->errors >= num_cases - best_class_count - Epsilon ) {
	(*tree)->leaf=1;
	(*tree)->num_children=0;
      }
    }
  }
}

int best_C45(int * used_examples, int * used_attr, float * split_val, float * inf, float * avg_distance)
{
  (*split_val)=0.0;
  int i,j,k,l,best=-1,n,uniques=0;
  struct continuous_sort * tmp;
  float float_best=0.0, tmp_distance, base_distance=0.0, pot_distance, *distance, *gain, 
    *split_vals, val, *thresholds, *unq, *Q, **M , **N,count=0.0, *left, *right;


  distance      =malloc(num_attrs*sizeof(float));
  gain          =malloc(num_attrs*sizeof(float));
  split_vals    =malloc(num_attrs*sizeof(float));
  thresholds    =malloc(num_attrs*sizeof(float));
  unq           =malloc(num_attrs*sizeof(float));

  left          =malloc(num_attrs*sizeof(float));
  right         =malloc(num_attrs*sizeof(float));

  for(i=0;i<num_attrs;i++) {
    distance[i]   =0.0;
    gain[i]       =0.0;
    split_vals[i] =0.0;
    inf[i]        =0.0;
    thresholds[i] =0.0;
    unq[i]        =0.0;
    left[i]       =0;
    right[i]      =0;
  }

  for(j=0;j<num_train;j++) {
    if(used_examples[j]==0) {
      count+=1.0;
    }
  }

  if(infgain_dist==1) {
    Q=malloc(num_classes*sizeof(float));
    //N=malloc(num_classes*sizeof(int *));
    for(j=0;j<num_classes;j++) {
      //N[j]=malloc(1*sizeof(int));
      //N[j][0]=0;
      Q[j]=0;
    }
    for(j=0;j<num_train;j++)
      {
	if(used_examples[j]==0)
	  {
	    //N[train_class_list[j]][0]++;
	    Q[train_class_list[j]]+=1.0;
	  }
	//base_distance = calcdistance(N,num_classes,1,used_examples);
	base_distance = totalgain2(Q, num_classes) / count;
      }
    //for(j=0;j<num_classes;j++) {
      //free(N[j]);
      //}
    //free(N);
    free(Q);
  }

  for(i=0;i<num_attrs;i++) {
    if(used_attr[i]==0) {
      if(attr_types[i]==CONTINUOUS) {
	int n_splits, * Y;
	float * splits, * X;

	X = malloc(count*sizeof(float));
	Y = malloc(count*sizeof(int));

	k=0;
	for(j=0;j<num_train;j++)
	  {
	    if(used_examples[j]==0)
	      {
		X[k]=train_continuous_list[j][i];
		Y[k]=train_class_list[j];
		k++;
	      }
	  }
	uniques=0;
	n_splits=find_splits(X, Y, count, &splits, &uniques);
	if(n_splits>0) {
	  N=malloc(num_classes*sizeof(float *));
	  M   =malloc(2*sizeof(float *));
	  M[0]=malloc(1*sizeof(float));
	  M[1]=malloc(1*sizeof(float));
	  for(j=0;j<num_classes;j++) {
	    N[j]=malloc(2*sizeof(float));
	    N[j][0]=0.0;
	    N[j][1]=0.0;
	  }
	  for(k=0;k<n_splits;k++) {
	    int n_left=0,n_right=0;
	    M[0][0]=0.0;
	    M[1][0]=0.0;
	    for(l=0;l<num_classes;l++) {
	      N[l][0]=0.0;
	      N[l][1]=0.0;
	    }
	    for(j=0;j<count;j++) {
	      if(X[j] <= splits[k]) {
		N[Y[j]][0]+=1.0;
		n_left++;
		M[0][0]+=1.0;
	      }
	      else {
		N[Y[j]][1]+=1.0;
		n_right++;
		M[1][0]+=1.0;
	      }
	    }
	    if(purity==1) {
               if (infgain_dist==1) {
	          tmp_distance=computegain(N, base_distance, 0.0, 2, count);
	          pot_distance=totalgain(M,2,1);
 	          if(tmp_distance > distance[i]) {
		     distance[i]   =tmp_distance;
		     gain[i]       =tmp_distance;
		     inf[i]        =pot_distance;
 	             split_vals[i] =splits[k];
		     left[i]       =M[0][0];
	             right[i]      =M[1][0];
 	          }
	          thresholds[i]=Log((double) uniques)/count;
	          unq[i]=uniques;
               }
               if (gini_dist==1) {
                  tmp_distance=gini(N, num_classes,2);
                  if(tmp_distance > distance[i]) {
                     distance[i]   =tmp_distance;
                     split_vals[i] =splits[k];
                     left[i]       =M[0][0];
                     right[i]      =M[1][0];
                  }
               }
               if(gini_skewfree_dist==1) {
                  tmp_distance=gini_skewfree(N, num_classes,2);
                  if(tmp_distance > distance[i]) {
                     distance[i]   =tmp_distance;
                     split_vals[i] =splits[k];
                     left[i]       =M[0][0];
                     right[i]      =M[1][0];
                  }
               }
               if (dkm_dist==1) {
                  tmp_distance=dkm(N,num_classes,0,1);
                  if(tmp_distance > distance[i]) {
                     distance[i]   =tmp_distance;
                     split_vals[i] =splits[k];
                     left[i]       =M[0][0];
                     right[i]      =M[1][0];
                  }
               }
               if (dkm_avg_dist==1) {
                  tmp_distance=dkm_avg(N,num_classes,0,1);
                  if(tmp_distance > distance[i]) {
                     distance[i]   =tmp_distance;
                     split_vals[i] =splits[k];
                     left[i]       =M[0][0];
                     right[i]      =M[1][0];
                  }
               }
               if (dkm_1againstmax_dist==1) {
                  tmp_distance=dkm_1against_max(N,num_classes,0,1);
                  if(tmp_distance > distance[i]) {
                     distance[i]   =tmp_distance;
                     split_vals[i] =splits[k];
                     left[i]       =M[0][0];
                     right[i]      =M[1][0];
                  }
               }
               if (dkm_1againstavg_dist==1) {
                  tmp_distance=dkm_1against_avg(N,num_classes,0,1);
                  if(tmp_distance > distance[i]) {
                     distance[i]   =tmp_distance;
                     split_vals[i] =splits[k];
                     left[i]       =M[0][0];
                     right[i]      =M[1][0];
                  }
               }
	    }
	    else {
	      tmp_distance=calcdistance(N,num_classes,2,used_examples);
	      if(tmp_distance > distance[i]) {
		distance[i]=tmp_distance;
		split_vals[i] =splits[k];
		left[i]       =M[0][0];
		right[i]      =M[1][0];
	      }
	    }
	  }
	  //thresholds[i]=uniques;
	  free(splits);
	  for(j=0;j<num_classes;j++) {
	    free(N[j]);
	  }
	  for(j=0;j<2;j++) {
	    free(M[j]);
	  }
	  free(N);
	  free(M);
	}
      }
      else if (attr_types[i]==NOMINAL){
	N=malloc(num_classes*sizeof(int *));
        M=malloc(num_attr_vals[i]*sizeof(float *));

        for(j=0;j<num_attr_vals[i];j++) {
           M[j]=malloc(1*sizeof(float));
        }
	for(j=0;j<num_classes;j++) {
	  N[j]=malloc(num_attr_vals[i]*sizeof(int));
	  for(k=0;k<num_attr_vals[i];k++) {
	    N[j][k]=0;
            M[k][0]=0;
	  }
	}
      
	for(j=0;j<num_train;j++) {
	  if(used_examples[j]==0) {
	    N[train_class_list[j]][train_attr_list[j][i]]++;
            M[train_attr_list[j][i]][0]++;
	  }
	}
	if(infgain_dist==1) {
           tmp_distance=computegain(N, base_distance, 0.0, num_attr_vals[i], count);
           pot_distance=totalgain(M,num_attr_vals[i],1);
           if(tmp_distance > distance[i]) {
              distance[i]   =tmp_distance;
              gain[i]       =tmp_distance;
              inf[i]        =pot_distance;
           }
           thresholds[i]=Log((double) uniques)/count;
           unq[i]=uniques;
	}
        else if (gini_dist==1) {
          distance[i]=gini(N, num_classes,num_attr_vals[i]);
        }
        else if (gini_skewfree_dist==1) {
          distance[i]=gini_skewfree(N, num_classes,num_attr_vals[i]);
        }
        else if (dkm_dist==1) {
          distance[i]=dkm1(N, num_classes,num_attr_vals[i]);
          /*
           int count=0;
           tmp_distance=0;
           for(j=1;j<num_attr_vals[i];j++) {
              for(k=0;k<j;k++) {
                 tmp_distance+=dkm(N, num_classes,k,j);
                 count++;
              }
           }
           distance[i]=(1.0*tmp_distance/count);*/
        }
        else if(dkm_avg_dist==1) {
           distance[i]=dkm1_avg(N, num_classes,num_attr_vals[i]);
        }
        else if(dkm_1againstmax_dist==1) {
           distance[i]=dkm1_1against_max(N, num_classes,num_attr_vals[i]);
        }
        else if(dkm_1againstavg_dist==1) {
           distance[i]=dkm1_1against_avg(N, num_classes,num_attr_vals[i]);
        }
	else {
	  distance[i]=calcdistance(N,num_classes,num_attr_vals[i],used_examples);
	}
	for(j=0;j<num_classes;j++) {
	  free(N[j]);
	}
        for(j=0;j<num_attr_vals[i];j++) {
          free(M[j]);           
        }
        free(M);
	free(N);
      }
    }
  }
  if(infgain_dist==1) {
    j=0;
    (*avg_distance)=0;
    for(i=0;i<num_attrs;i++) {
      gain[i]=gain[i]-thresholds[i];
      if(gain[i]>-Epsilon) {
	(*avg_distance)+=gain[i];
	j++;
      }
    }
    if(j>0) {
      (*avg_distance) = (*avg_distance) / j;
    }
    
    for(i=0;i<num_attrs;i++) {
      val=Worth(inf[i],gain[i],(*avg_distance));
      if(val>float_best) {
	best=i;
	float_best=val;
	(*split_val)=split_vals[i];
      }
    }
  }
  else {
    for(i=0;i<num_attrs;i++) {
      if(distance[i] > float_best) {
	float_best=distance[i];
	best=i;
	if(attr_types[i]==CONTINUOUS) {
	  (*split_val)=split_vals[i];
	}
      }
    }
  }
  free(distance);
  free(split_vals);
  //printf("highest: %f, returning best %d\n", float_best, best);
  return best;
}

float Worth(float ThisInfo, float ThisGain, float MinGain)
{
  if ( GAINRATIO == 1 ) {
    if ( ThisGain - MinGain >= -Epsilon && ThisInfo > Epsilon ) {
      return ThisGain / ThisInfo;
    } else {
      return -Epsilon;
    }
  } else
    return ( ThisInfo > 0 && ThisGain > -Epsilon ? ThisGain : -Epsilon );
}

float calcdistance(int ** N, int m, int n, int * used_examples)
{
  float distance;
  int j,k,count,a,b;
  float tmpa,tmpb;
  if(euclidean_dist==1) {
    distance=euclidean(N,m,n);
  }
  else if(chord_dist==1) {
    distance=chord(N,m,n,0,1);
  }
  else if(chisquaremetric_dist==1) {
    distance=chisquaremetric(N,m,n,0,1);
  }
  else if(chisquaredistance_dist==1) {
    distance=chisquaredistance(N,m,n,0,1);
  }
  else if(speciesprofile_dist==1) {
    distance=speciesprofile(N,m,n,0,1);
  }
  else if(hellinger_dist==1) {
   if(m==2) {
      distance=hellinger1(N,m,n,0,1);
   }
   else {
      distance=0;
      count=0;
      for(b=1;b<m;b++) {
         for(a=0;a<b;a++) {
            tmpa=0,tmpb=0;
            for(j=0;j<n;j++) {
               tmpa+=N[a][j];
               tmpb+=N[b][j];
            }
            if((tmpa > 0) && (tmpb > 0)) {
               distance+=hellinger1(N,m,n,a,b);
               count++;
            }
         }
      }
      distance=distance/count;
      //distance=hellinger(N,m,n);
   }
  }
  else if(hellinger_1againstmax_dist==1) {
    distance=hellinger_1against_max(N,m,n);
  }
  else if(hellinger_weightoutside_dist==1) {
     distance=hellinger1_weightoutside(N,m,n,0,1);
  }
  else if(hellinger_weightinside_dist==1) {
     distance=hellinger1_weightinside(N,m,n,0,1);
  }
  else if(hellinger_weightboth_dist==1) {
     distance=hellinger1_weightboth(N,m,n,0,1);
  }
  else if(hellinger_inv_dist==1) {
    if(m==2) {
       distance=hellinger_inv1(N,m,n,0,1);
    }
    else {
       distance=hellinger_inv(N,m,n);
    }
  }
  else { //infgain
    /*int ** N_y;
    N_y=malloc(sizeof(int *)*m);
    for(k=0;k<m;k++) {
      N_y[k]=malloc(sizeof(int)*1);
      N_y[k][0]=0;
    }
    for(k=0;k<num_train;k++) {
      if(used_examples[k]==0) {
	N_y[train_class_list[k]][0]++;
      }
    }
    distance=infgain(N_y,m,1) - infgain(N,m,n);
    for(k=0;k<m;k++) {
      free(N_y[k]);
    }
    free(N_y);*/
    distance=infgain(N,m,n);
  }
  return distance;
}
