/*************************************************************************/
/*									 */
/*	Evaluation of a test on a discrete valued attribute		 */
/*      ---------------------------------------------------		 */
/*									 */
/*************************************************************************/
#include "buildex.h"


/*************************************************************************/
/*									 */
/*  Set Info[] and Gain[] for discrete partition of items Fp to Lp	 */
/*									 */
/*************************************************************************/
void EvalDiscreteAtt(int a,int Fp,int Lp,double number_of_items)
{ 
  float known_items;
  
  compute_frequencies(a, Fp, Lp);
  
  known_items = number_of_items - ValFreq[0];
  
  /*  Special case when no known values of the attribute  */
  if ( number_of_items <= ValFreq[0] ) {
    Verbosity(2) printf("\tAtt %s: no known values\n", attribute_name[a]);
    Gain[a] = -Epsilon;
    Info[a] = 0.0;
    return;
  }
  
  Gain[a] = ComputeGain(DiscrKnownBaseInfo(known_items, MaxAttVal[a]),
			  UnknownRate[a], MaxAttVal[a], known_items);
  Info[a] = TotalInfo(ValFreq, 0, MaxAttVal[a]+1) / number_of_items;
  
  Verbosity(2) {
    printf("\tAtt %s", attribute_name[a]);
    Verbosity(3) PrintDistribution(a, MaxAttVal[a], True);
    printf("\tinf %.3f, gain %.3f\n", Info[a], Gain[a]);
  }
  
  return;
} 



/*************************************************************************/
/*									 */
/*  Compute frequency tables Freq[][] and ValFreq[] for attribute	 */
/*  Att from items Fp to Lp, and set the UnknownRate for a		 */
/*									 */
/*************************************************************************/
void compute_frequencies(int a, int Fp, int Lp)
{
  Description Case; 
  int c,p,v;
  
  reset_frequencies(MaxAttVal[a]);
  
  /*  Determine the frequency of each class amongst cases
      with each possible value for the given attribute  */
  for (p=Fp; p <= Lp; p++) {
    Case = Item[p];
    Freq[ DVal(Case,a) ][ Class(Case) ] += Weight[p];
  } 
  
  /*  Determine the frequency of each possible value for the
      given attribute  */
  for (v=0; v <= MaxAttVal[a];v++) {
    for (c=0; c < number_of_classes; c++) {
      ValFreq[v] += Freq[v][c];
    }
  }
  
  /*  Set the rate of unknown values of the attribute  */
  UnknownRate[a] = ValFreq[0] / CountItems(Fp, Lp);
  
  return;
}



/*************************************************************************/
/*									 */
/*  Return the base info for items with known values of a discrete	 */
/*  attribute, using the frequency table Freq[][]			 */
/*	 								 */
/*************************************************************************/
float DiscrKnownBaseInfo(ItemCount KnownItems, DiscrValue MaxVal)
{
  int c,v;
  ItemCount ClassCount;
  double Sum=0;
  
  for (c=0; c < number_of_classes; c++) {
    ClassCount = 0;
    for (v=1; v <= MaxVal; v++) 
      ClassCount += Freq[v][c];
    Sum += ClassCount * Log(ClassCount);
  }
  
  return (KnownItems * Log(KnownItems) - Sum) / KnownItems;
}



/*************************************************************************/
/*									 */
/*  Construct and return a node for a test on a discrete attribute	 */
/*									 */
/*************************************************************************/
void DiscreteTest(TREE *node, int a)
{
  
  Sprout(node, MaxAttVal[a]);
  
  node->node_type=DISCRETE_NODE;
  node->tested_attribute=a;
  node->Errors	= 0;
  
  return;
}

