#include "distance.h"


/*
 *
 * These distance functions taken from the Legendre and Gallagher
 * "Ecoloically meaningful transformations for ordination of species
 * data" paper
 *
 */

//calculates the euclidean between rows j1 and j2 of the {m x n} array y
float euclidean(float ** y, int m, int n) {    
  int i,j,a,c;
  int * y_plus;
  float result=0.0;
  float final=0.0;
  
  y_plus=malloc(sizeof(int)*m);

  for(j=0;j<m;j++) {
     y_plus[j]=0;
  }

  for(i=0;i<n;i++) {
     for(j=0;j<m;j++) {
        y_plus[j]+=y[j][i]*y[j][i];
     }
  }
  c=0;
  for(a=1;a<m;a++) {
     for(j=0;j<a;j++) {
        result=0.0;
        c++;
        for(i=0;i<n;i++) { 
           result+= pow( 1.0*y[a][i] - 1.0*y[j][i], 2.0);
        }
        final+=sqrt(result);
     }
  } 
  
  return final/c;
} 


//calculates the euclidean between rows j1 and j2 of the {m x n} array y
/*float euclidean(float ** y, int m, int n, int j1, int j2) {

  int i;
  int y_plus[2];
  float result=0.0;

  if(j1 < 0 || j2 < 0 || j1 > m || j2 > m)
    {
      return -1.0;
    }

  y_plus[0]=0;
  y_plus[1]=0;
  for(i=0;i<n;i++)
    {
      y_plus[0]+=y[j1][i]*y[j1][i];
      y_plus[1]+=y[j2][i]*y[j2][i];
    }

  for(i=0;i<n;i++)
    {
      result+= pow( 1.0*y[j1][i] - 1.0*y[j2][i], 2.0);
    }

  return sqrt(result);
}
*/
//calculates the euclidean distance between rows j1 and j2 of the {m x n} array y
float chord(float ** y, int m, int n, int j1, int j2) {

  int i;
  int y_plus[2];
  float result=0.0;

  if(j1 < 0 || j2 < 0 || j1 > m || j2 > m)
    {
      return -1.0;
    }

  y_plus[0]=0;
  y_plus[1]=0;
  for(i=0;i<n;i++)
    {
      y_plus[0]+=y[j1][i]*y[j1][i];
      y_plus[1]+=y[j2][i]*y[j2][i];
    }

  for(i=0;i<n;i++)
    {
      result+= pow( 1.0*y[j1][i]/sqrt(1.0*y_plus[0]) - 1.0*y[j2][i]/sqrt(1.0*y_plus[1]), 2.0);
    }

  return sqrt(result);
}

//calculates the chi-square metric between rows j1 and j2 of the {m x n} array y
float chisquaremetric(float ** y, int m, int n, int j1, int j2) {

  int i,j;
  int y_plus[2];
  int * x_plus;
  float result=0.0;

  if(j1 < 0 || j2 < 0 || j1 > m || j2 > m)
    {
      return -1.0;
    }

  x_plus=malloc(sizeof(int)*m);
  for(i=0;i<m;i++)
    {
      x_plus[i]=0;
    }

  for(i=0;i<m;i++)
    {
      for(j=0;j<n;j++)
	{
	  x_plus[i]+=y[j][i];
	}
    }

  y_plus[0]=0;
  y_plus[1]=0;
  for(i=0;i<n;i++)
    {
      y_plus[0]+=y[j1][i];
      y_plus[1]+=y[j2][i];
    }

  for(i=0;i<n;i++)
    {
      result+=1.0/(1.0*x_plus[i])*pow(1.0*y[j1][i]/y_plus[0] - 1.0*y[j2][i]/y_plus[1], 2.0);
    }
  free(x_plus);
  return sqrt(result);
}

//calculates the chi-square distance between rows j1 and j2 of the {m x n} array y
float chisquaredistance(float ** y, int m, int n, int j1, int j2) {

  int i,j,y_plus_plus=0;
  int y_plus[2];
  int * x_plus;
  float result=0.0;

  if(j1 < 0 || j2 < 0 || j1 > m || j2 > m)
    {
      return -1.0;
    }

  x_plus=malloc(sizeof(int)*m);
  for(i=0;i<m;i++)
    {
      x_plus[i]=0;
    }

  for(i=0;i<m;i++)
    {
      for(j=0;j<n;j++)
	{
	  x_plus[i]+=y[j][i];
	  y_plus_plus+=y[j][i];
	}
    }

  y_plus[0]=0;
  y_plus[1]=0;
  for(i=0;i<n;i++)
    {
      y_plus[0]+=y[j1][i];
      y_plus[1]+=y[j2][i];
    }

  for(i=0;i<n;i++)
    {
      result+=(1.0*y_plus_plus)/(1.0*x_plus[i])*pow(1.0*y[j1][i]/y_plus[0] - 1.0*y[j2][i]/y_plus[1], 2.0);
    }
  free(x_plus);
  return sqrt(result);
}

//calculates the species profile between rows j1 and j2 of the {m x n} array y
float speciesprofile(float ** y, int m, int n, int j1, int j2) {

  int i,j;
  int y_plus[2];
  float result=0.0;

  if(j1 < 0 || j2 < 0 || j1 > m || j2 > m)
    {
      return -1.0;
    }

  y_plus[0]=0;
  y_plus[1]=0;
  for(i=0;i<n;i++)
    {
      y_plus[0]+=y[j1][i];
      y_plus[1]+=y[j2][i];
    }

  for(i=0;i<n;i++)
    {
      result+=pow(1.0*y[j1][i]/y_plus[0] - 1.0*y[j2][i]/y_plus[1],2.0);
    }
  return sqrt(result);
}

//calculates the hellinger distance between all rows of the {m x n} array y
float hellinger(float ** y, int m, int n) {
   int i,j,k,c;
   int count;
   int * y_plus;
   float result=0.0,row,tmp;

   y_plus=malloc(m*sizeof(int));

   for(j=0;j<m;j++) {
      y_plus[j]=0;
   }

   for(i=0;i<n;i++) {
      for(j=0;j<m;j++) {
         y_plus[j]+=y[j][i];
      }
   }
   c = m*(m-1)/2;
   //tmp = malloc(sizeof(float *)*n);
   //for(i=0;i<n;i++) {
   //   tmp[i] =  malloc(sizeof(float)*c);
   //}

   for(i=0;i<n;i++) {
      c=0;
      count=0;
      tmp=0;
      for(j=0;j<m-1;j++) {
         for(k=j+1;k<m;k++) {
            if( y_plus[j]!=0 && y_plus[k]!=0 ) {
               tmp+=pow( sqrt( (1.0*y[j][i]/(1.0*y_plus[j]))) - sqrt( (1.0*y[k][i]/(1.0*y_plus[k]))), 2.0);
               c++;
               count++;
            }
         }
      }
      result+=sqrt(tmp)/count;
   }
   /**c = m*(m-1)/2;
   for(j=0;j<c;j++) {
      row=0.0;
      for(i=0;i<n;i++) {
         row+=tmp[i][c];
      }
      result+=sqrt(row)/c;
   }

   for(i=0;i<n;i++) {
      free(tmp[i]);
   }*/

   free(y_plus);
   //free(tmp);
   return (result);
}

//avg and not max, but I am saving time
float hellinger_1against_max(float ** y, int m, int n) {
   int i,j,k;
   int y_plus[2];
   float result=0.0,tmp,final=0.0;

   for(k=0;k<m;k++) {
      y_plus[0]=0;
      y_plus[1]=0;

      for(i=0;i<n;i++) {
         y_plus[0]+=y[k][i];
         for(j=0;j<m;j++) {
            if(j!=k) {
               y_plus[1]+=y[j][i];
            }
         }
      }

      result=0.0;
      for(i=0;i<n;i++) {
         tmp=0.0;
         for(j=0;j<m;j++) {
            if(j!=k) {
               tmp+=y[j][i];
            }
         }
         if( y_plus[0]!=0 && y_plus[1]!=0 ) {
            result+=pow( sqrt( (1.0*y[k][i]/(1.0*y_plus[0]))) - sqrt( (1.0*tmp/(1.0*y_plus[1]))), 2.0);
         }
      }
      result=sqrt(result);
      if(result>final) {
         final=result;
      }
   }

   return (final);
}

float hellinger1(float ** y, int m, int n, int y1, int y2) {
   int i;
   int y_plus[2];
   float result=0.0;

   y_plus[0]=0;
   y_plus[1]=0;

   for(i=0;i<n;i++) {
      y_plus[0]+=y[y1][i];
      y_plus[1]+=y[y2][i];
   }

   for(i=0;i<n;i++) {
      if( y_plus[0]!=0 && y_plus[0]!=0 ) {
         result+=pow( sqrt( (1.0*y[y1][i]/(1.0*y_plus[0]))) - sqrt( (1.0*y[y2][i]/(1.0*y_plus[1]))), 2.0);
      }
   }

   return (sqrt(result));
}

float hellinger1_weightoutside(float ** y, int m, int n, int y1, int y2) {
   int i;
   int y_plus[2];
   float result=0.0;
   int total;

   y_plus[0]=0;
   y_plus[1]=0;

   for(i=0;i<n;i++) {
      y_plus[0]+=y[y1][i];
      y_plus[1]+=y[y2][i];
      total+=y[y1][i];
      total+=y[y2][i];
   }

   for(i=0;i<n;i++) {
      if( y_plus[y1]!=0 && y_plus[y2]!=0 ) {
         result+=(1.0*(y[y1][i]+y[y2][i])/total)*pow( sqrt( (1.0*y[y1][i]/(1.0*y_plus[0]))) - sqrt( (1.0*y[y2][i]/(1.0*y_plus[1]))), 2.0);
      }
   }

   return (sqrt(result));
}

float hellinger1_weightinside(float ** y, int m, int n, int y1, int y2) {
   int i;
   int y_plus[2];
   float result=0.0;

   y_plus[0]=0;
   y_plus[1]=0;

   for(i=0;i<n;i++) {
      y_plus[0]+=y[y1][i];
      y_plus[1]+=y[y2][i];
   }

   for(i=0;i<n;i++) {
      if( y_plus[y1]!=0 && y_plus[y2]!=0 ) {
         result+=pow( (1.0*y[y1][i]/(y[y1][i]+y[y2][i]))*sqrt( (1.0*y[y1][i]/(1.0*y_plus[0]))) - (1.0*y[y2][i]/(y[y1][i]+y[y2][i]))*sqrt( (1.0*y[y2][i]/(1.0*y_plus[1]))), 2.0);
      }
   }

   return (sqrt(result));
}


float hellinger1_weightboth(float ** y, int m, int n, int y1, int y2) {
   int i;
   int y_plus[2],total;
   float result=0.0;

   y_plus[0]=0;
   y_plus[1]=0;

   for(i=0;i<n;i++) {
      y_plus[0]+=y[y1][i];
      y_plus[1]+=y[y2][i];
      total+=y[y1][i];
      total+=y[y2][i];
   }

   for(i=0;i<n;i++) {
      if( y_plus[y1]!=0 && y_plus[y2]!=0 ) {
         result+=(1.0*(y[y1][i]+y[y2][i])/total)*pow( (1.0*y[y1][i]/(y[y1][i]+y[y2][i]))*sqrt( (1.0*y[y1][i]/(1.0*y_plus[0]))) - (1.0*y[y2][i]/(y[y1][i]+y[y2][i]))*sqrt( (1.0*y[y2][i]/(1.0*y_plus[1]))), 2.0);
      }
   }

   return (sqrt(result));
}

//calculates the hellinger_inv distance between rows j1 and j2 of the {m x n} array y
float hellinger_inv(float ** y, int m, int n) {

  int i,j,k,c;
  float * y_plus;
  float result=0.0,** tmp,row;

  y_plus=malloc(sizeof(float)*n);

  for(j=0;j<n;j++) {
     y_plus[j]=0.0;
  }
  for(i=0;i<m;i++) {
     for(j=0;j<n;j++) {
        y_plus[j]+=y[i][j];
     }
  }

  tmp=malloc(sizeof(float)*m);
  c=n*(n-1)/2;
   for(i=0;i<m;i++) {
      tmp[i] =  malloc(sizeof(float)*c);
   }

  for(i=0;i<m;i++) {
     c=0;
     for(j=0;j<n-1;j++) {
        for(k=j+1;k<n;k++) {
           if( y_plus[j]!=0 && y_plus[k]!=0 ) {
              tmp[i][c] = pow( sqrt( (1.0*y[i][j]/(1.0*y_plus[j]))) - sqrt( (1.0*y[i][k]/(1.0*y_plus[k]))), 2.0);
              c++;
           }
        }
     }
  }

   c = n*(n-1)/2;
   for(j=0;j<c;j++) {
      row=0.0;
      for(i=0;i<m;i++) {
         row+=tmp[i][c];
      }
      result+=sqrt(row)/c;
   }

   for(i=0;i<m;i++) {
      free(tmp[i]);
   }

  free(tmp);
  free(y_plus);
  return result;
}

float hellinger_inv1(float ** y, int m, int n, int y1, int y2) {

  int i;
  float y_plus[2];
  float result=0.0;

  y_plus[0]=0.0;
  y_plus[1]=0.0;

  for(i=0;i<m;i++) {
     y_plus[0]+=y[i][y1];
     y_plus[1]+=y[i][y2];
  }

  for(i=0;i<m;i++) {
     if( y_plus[y1]!=0 && y_plus[y2]!=0 ) {
        result+=pow( sqrt( (1.0*y[i][y1]/(1.0*y_plus[0]))) - sqrt( (1.0*y[i][y1]/(1.0*y_plus[0]))), 2.0);
     }
  }

  return (sqrt(result));
}

//calculates the information gain of the {m x n} array y
float infgain(int ** y, int m, int n)
{
  int i,j;
  float y_plus_plus=0.0, *y_plus;
  float result=0.0;

  y_plus=malloc(sizeof(int)*n);
  for(i=0;i<n;i++)
    {
      y_plus[i]=0.0;
    }

  for(i=0;i<m;i++)
    {
      for(j=0;j<n;j++)
	{	  
	  y_plus[j]+=y[i][j];
	  y_plus_plus+=y[i][j];
	}
    }

  for(i=0;i<m;i++)
    {
      for(j=0;j<n;j++)
	{
	  result-=(y[i][j])/(y_plus_plus)*Log( y[i][j]/y_plus[j] );
	}
    }
  
  free(y_plus);
  return result;
}

//calculates the information gain of the {m x n} array y
float totalgain(float ** y, int m, int n)
{
  int i,j;
  float result=0.0, total=0.0;

  for(i=0;i<m;i++)
    {
      for(j=0;j<n;j++)
	{
	  result +=y[i][j]*Log(y[i][j]);
	  total  +=y[i][j];
	}
    }
  return ( (total * Log(total) - result) / total);
}

float totalgain2(float * y, int m)
{
  int i;
  float result=0.0, total=0.0;

  for(i=0;i<m;i++)
    {
      result +=y[i]*Log(y[i]);
      total  +=y[i];
    }
  //printf("%f %f %f\n",total, Log(total), result);
  return ( total * Log(total) - result);
  //return ( (total * Log(total) - result) / total);
}

float computegain(float ** y, float BaseInfo, float UnknFrac, int m, float n)
{
  int v,w;
  float ThisInfo=0.0, ThisGain;
  short ReasonableSubsets=0;
  
  float * M;

  M=malloc(num_classes*sizeof(float *));
  for(w=0; w < num_classes; w++) {
    M[w]=0.0;
  }

  //m = number of splits
  //n = number of items in the set
  /*  Check whether all values are unknown or the same  */
  if (  n <= 0.0 ) return -Epsilon;
  
  /*  There must be at least two subsets with MINOBJS items  */

  for (v=0; v < m; v++) {
    float total=0.0;
    for(w=0; w < num_classes; w++) {
      total+=y[w][v];
    }
    if ( total >= MINOBJS ) ReasonableSubsets++;
  }
  if ( ReasonableSubsets < 2 ) return -Epsilon;
  
  /*  Compute total info after split, by summing the
      info of each of the subsets formed by the test  */
  
  for (v=0; v < m; v++) {
    for(w=0; w < num_classes; w++) {
      M[w] = y[w][v];
      //printf("M[%d] = %f\n",w,M[w]);
    }
    //printf("%f\n",totalgain2(M, num_classes));
    ThisInfo += totalgain2(M, num_classes);
  }
  
  
  /*  Set the gain in information for all items, adjusted for unknowns  */
  ThisGain = (1 - UnknFrac) * (BaseInfo - ThisInfo / n);
    
  //printf("%f %f %f %f %f\n", ThisGain, UnknFrac, BaseInfo, ThisInfo, n);

  free(M);
  return ThisGain;
}


/* Linux/Solaris interoperability issues required this to be a
   function. For some reason, using a macro and casting the 
   result as a float did not work. The logs would be used as
   doubles, and Linux/Solaris differ at double precision on the function */
//
// float Log(double d)
//{
//   float retval=0;
//   static const float log2=0.69314718055994530942f;  /*From Quinlan, defns.i */

//   if ( d < 1E-6) return 0;
   
//   retval=((float)log(d))/log2;
//   return retval;
//}



float Log(double d)
{
  float retval;
  if ( d < 1E-6) return 0;

  retval=log(d)/log(2.0);
  return retval;
}

float gini(float ** y, int m, int n) {
   int i,j,count;
   int * y_plus; 
   float *y_class, total;
   float result=0.0,tmp,denom,numer,base_gini,val;
   float * split;
   denom=0.0;

   y_plus=malloc(n*sizeof(int));
   y_class=malloc(m*sizeof(float));
   split=malloc(n*sizeof(float));
   for(j=0;j<n;j++) {
      y_plus[j]=0;
      split[j]=0.0;
   }
   for(i=0;i<m;i++) {
      y_class[i]=0;
   }
   

   for(i=0;i<n;i++) {
      for(j=0;j<m;j++) {
         y_plus[i]+=y[j][i];
         y_class[j]+=y[j][i];
         denom+=y[j][i];
         total+=y[j][i];
      }
   }
 
   tmp=0.0;
   for(j=0;j<m;j++) {
      val = (1.0*y_class[j]/total)*(1.0*y_class[j]/total);
      tmp +=(1.0*y_class[j]/total)*(1.0*y_class[j]/total);
   }
   base_gini=1-tmp;

   for(i=0;i<n;i++) {
      tmp=0.0;
      count=0;
      for(j=0;j<m;j++) {
         if( y_plus[i]!=0 ) {
            split[i]+=(1.0*y[j][i]/y_plus[i])*(1.0*y[j][i]/y_plus[i]);
         }
      }
      split[i]=1-split[i];
   }
   numer=0;
   for(i=0;i<n;i++) {
      val   =y_plus[i]*split[i];
      numer+=y_plus[i]*split[i];
   }
   free(y_plus);
   free(split);
   free(y_class);

   return (base_gini-numer/denom);
}


float gini_skewfree(float ** y, int m, int n) {
   int i,j,count;
   int * y_plus; 
   float *y_class, total;
   float result=0.0,tmp,denom,numer,base_gini,val;
   float * split;
   denom=0.0;

   y_plus=malloc(n*sizeof(int));
   y_class=malloc(m*sizeof(float));
   split=malloc(n*sizeof(float));
   for(j=0;j<n;j++) {
      y_plus[j]=0;
      split[j]=0.0;
   }
   for(i=0;i<m;i++) {
      y_class[i]=0;
   }
   

   for(i=0;i<n;i++) {
      for(j=0;j<m;j++) {
         y_plus[i]+=y[j][i];
         y_class[j]+=y[j][i];
         total+=y[j][i];
      }
   }
  
   base_gini=1;

   for(i=0;i<n;i++) {
      tmp=0.0;
      denom=0.0;
      split[i]=2.0;
      for(j=0;j<m;j++) {
         if( y_class[i]!=0 ) {
            //split[i]+=(1.0*y[j][i]/y_plus[i])*(1.0*y[j][i]/y_plus[i]);
            split[i]*=y[j][i]/y_class[j];
            denom   +=y[j][i]/y_class[j];
         }
      }
      split[i]=split[i]/denom;
   }
   result=0.0;
   for(i=0;i<n;i++) {
      result+=split[i];
   }
   free(y_plus);
   free(split);
   free(y_class);

   return (base_gini-result);
}

float dkm(float ** y, int m, int y1, int y2) {
   int i;
   float result =0.0;
   float p_left =0.0;
   float p_right=0.0;
   float n_left =0.0;
   float n_right=0.0;
   float p_left_class=0.0;
   float p_right_class=0.0;
   float total=0.0;
   float left, right;
   float base=1.0;
   int * present;

   present=malloc(sizeof(int)*m);
   
   for(i=0;i<m;i++) {
      n_left +=y[i][y1];
      n_right+=y[i][y2];
      total+=y[i][y1];
      total+=y[i][y2];
      if((y[i][y1]+y[i][y2])>0) {
         present[i]=1;
      }
      else {
         present[i]=0;
      }
   }
   p_left =n_left/total;
   p_right=n_right/total;
   
   for(i=0;i<m;i++) {
      if( present[i]==1 ) {
         base=base*((y[i][y1]+y[i][y2])/total);
      }
   }
   base=2*sqrt(base);
   left=1.0;
   for(i=0;i<m;i++) {
       if( present[i]==1 ) {
          left = left*(y[i][y1]/n_left);
       }
   }
   left=2*sqrt(left);

   right=1.0;
   for(i=0;i<m;i++) {
       if( present[i]==1 ) {
          right = right*(y[i][y2]/n_right);
       }
   }
   right=2*sqrt(right);
  
   free(present);
   result=(n_left*left + n_right*right)/(n_left+n_right); 
   return (base-result);
}


float dkm_avg(float ** y, int m, int y1, int y2) {
   int i,j,k;
   float result =0.0, final=0.0;
   float p_left =0.0;
   float p_right=0.0;
   float n_left =0.0;
   float n_right=0.0;
   float p_left_class=0.0;
   float p_right_class=0.0;
   float total=0.0;
   float left, right;
   float base=1.0;
   int count=0;

   for(k=1;k<m;k++) {
      for(j=0;j<k;j++) {
         if(((y[j][y1]+y[j][y2]) > 0) && ((y[k][y1]+y[k][y2]) > 0)) {
            n_left  = y[j][y1]+y[k][y1];
            n_right = y[j][y2]+y[k][y2];
            total   = n_left + n_right;
            p_left  = n_left/total;
            p_right = n_right/total;
   
            base =2*sqrt(((y[j][y1]+y[j][y2])/total)*((y[k][y1]+y[k][y2])/total));
            left =2*sqrt((y[j][y1]/n_left)*(y[k][y1]/n_left));
            right=2*sqrt((y[j][y2]/n_right)*(y[k][y2]/n_right));
  
            result=(n_left*left + n_right*right)/(n_left+n_right);
            final+=(base-result);
            count++;
         }
      }
   }
   return (final/count);
}

float dkm_1against_avg(float ** y, int m, int y1, int y2) {
   int i,j,k;
   float result =0.0, final=0.0;
   float p_left =0.0;
   float p_right=0.0;
   float n_left =0.0;
   float n_right=0.0;
   float p_left_class=0.0;
   float p_right_class=0.0;
   float total=0.0;
   float left, right;
   float base=1.0;
   float one, many;

   for(k=0;k<m;k++) {
      n_left=0.0;
      n_right=0.0;
      total=0.0;
      for(i=0;i<m;i++) {
         n_left  += y[i][y1];
         n_right += y[i][y2];
         total   += y[i][y1] + y[i][y2];
      }
      p_left =n_left/total;
      p_right=n_right/total;
   
      one=y[k][y1]+y[k][y2];
      many=0.0;
      for(i=0;i<m;i++) {
         if(i!=k) {
            many+=y[i][y1]+y[i][y2];
         }
      }
      base =2*sqrt((one/total)*(many/total));
     
      one=y[k][y1];
      many=0.0;
      for(i=0;i<m;i++) {
         if(i!=k) {
            many+=y[i][y1];
         }
      }
      left =2*sqrt((one/n_left)*(many/n_left));

      one=y[k][y2];
      many=0.0;
      for(i=0;i<m;i++) {
         if(i!=k) {
            many+=y[i][y2];
         }
      }
      right=2*sqrt((one/n_right)*(many/n_right));
  
      result=(n_left*left + n_right*right)/(n_left+n_right);
      final+=base-result;      
   }
   return (final/m);
}


float dkm_1against_max(float ** y, int m, int y1, int y2) {
   int i,j,k;
   float result =0.0, final=0.0;
   float p_left =0.0;
   float p_right=0.0;
   float n_left =0.0;
   float n_right=0.0;
   float p_left_class=0.0;
   float p_right_class=0.0;
   float total=0.0;
   float left, right;
   float base=1.0;
   float one, many;

   for(k=0;k<m;k++) {
      n_left=0.0;
      n_right=0.0;
      total=0.0;
      for(i=0;i<m;i++) {
         n_left  += y[i][y1];
         n_right += y[i][y2];
         total   += y[i][y1] + y[i][y2];
      }
      p_left =n_left/total;
      p_right=n_right/total;
   
      one=y[k][y1]+y[k][y2];
      many=0.0;
      for(i=0;i<m;i++) {
         if(i!=k) {
            many+=y[i][y1]+y[i][y2];
         }
      }
      base =2*sqrt((one/total)*(many/total));
     
      one=y[k][y1];
      many=0.0;
      for(i=0;i<m;i++) {
         if(i!=k) {
            many+=y[i][y1];
         }
      }
      left =2*sqrt((one/n_left)*(many/n_left));

      one=y[k][y2];
      many=0.0;
      for(i=0;i<m;i++) {
         if(i!=k) {
            many+=y[i][y2];
         }
      }
      right=2*sqrt((one/n_right)*(many/n_right));
  
      result=(n_left*left + n_right*right)/(n_left+n_right);
      if((base-result) > final) {
         final=base-result;
      }
   }
   return (final);
}

float dkm1(float ** y, int m, int n) {
   int i, j;
   float tmp;
   float result =0.0;
   float * p;
   float * q;
   float * split;
   float total=0.0;
   float left, right;
   float base=1.0;
   int * present;

   p=malloc(n*sizeof(float));
   q=malloc(n*sizeof(float));  
   split=malloc(n*sizeof(float));
   present=malloc(m*sizeof(int));
 
   for(i=0;i<m;i++) {
      for(j=0;j<n;j++) {
         q[j] +=y[i][j];
         total+=y[i][j];
      }
   }
   for(j=0;j<n;j++) {
      p[j] =q[j]/total;
   }
   
   for(i=0;i<m;i++) {
      tmp=0.0;
      for(j=0;j<n;j++) {
         tmp+=y[i][j]/total;
      }
      if(tmp > 0) {
         base=base*tmp;
         present[i]=1;
      }
      else {
         present[i]=0;
      }
   }
   base=2*sqrt(base);
   //old method
   for(j=0;j<n;j++) {
      split[j]=1.0;
      for(i=0;i<m;i++) {
          if(present[i]==1) {
             split[j]= split[j] * y[i][j]/q[j];
          }
      }
      split[j]=2*sqrt(split[j]);
   }


   result=0.0;
   for(j=0;j<n;j++) {
      result+=q[j]*split[j];
   }

   free(p);
   free(q);
   free(split);
   free(present);
   return (base-result/total);
}

float dkm1_avg(float ** y, int m, int n) {
   int i, j;
   float tmp;
   float result =0.0;
   float * p;
   float * q;
   float * split;
   float total=0.0;
   float left, right;
   float base=1.0,final=0.0;
   int a,b;

   p=malloc(n*sizeof(float));
   q=malloc(n*sizeof(float));  
   split=malloc(n*sizeof(float));
 
   for(b=1;b<m;b++) {
      for(a=0;a<b;a++) {
         total=0;
         base =1.0;
         for(j=0;j<n;j++) {
            q[j]=0;
         }
         for(j=0;j<n;j++) {
            q[j] +=y[a][j]+y[b][j];
            total+=y[a][j]+y[b][j];
         }
   
         for(j=0;j<n;j++) {
            p[j] =q[j]/total;
         }
   
         tmp=0.0;
         for(j=0;j<n;j++) {
            tmp+=y[a][j]/total;
         }
         base=base*tmp;
         tmp=0.0;
         for(j=0;j<n;j++) {
            tmp+=y[b][j]/total;
         }
         base=base*tmp;
         base=2*sqrt(base);

         for(j=0;j<n;j++) {
            split[j]=1.0;
            split[j]= split[j]*y[a][j]/q[j]*y[b][j]/q[j];
            split[j]=2*sqrt(split[j]);
         }

         result=0.0;
         for(j=0;j<n;j++) {
            result+=q[j]*split[j];
         }
         final+=(base-result/total);
      }
   }   

   free(p);
   free(q);
   free(split);
   return (2*final/(m*(m-1)));
}

float dkm1_1against_avg(float ** y, int m, int n) {
   int i, j;
   float tmp;
   float result =0.0;
   float * p;
   float * q;
   float * split;
   float total=0.0;
   float left, right;
   float base=1.0,final=0.0;
   float one,many;
   int a;

   p=malloc(n*sizeof(float));
   q=malloc(n*sizeof(float));  
   split=malloc(n*sizeof(float));
 
   for(a=0;a<m;a++) {
      total=0;
      base =1.0;
      for(j=0;j<n;j++) {
         q[j]=0;
      }
      for(j=0;j<n;j++) {
         for(i=0;i<m;i++) {
            q[j] +=y[i][j];
            total+=y[i][j];
         }
      }
   
      for(j=0;j<n;j++) {
         p[j] =q[j]/total;
      }
   
      tmp=0.0;
      for(j=0;j<n;j++) {
         tmp+=y[a][j]/total;
      }
      base=base*tmp;
      tmp=0.0;
      for(j=0;j<n;j++) {
         for(i=0;i<m;i++) {
            if(i!=a) {
               tmp+=y[i][j]/total;
            }
         }
      }
      base=base*tmp;
      base=2*sqrt(base);

      for(j=0;j<n;j++) {
         split[j]=1.0;
         one=y[a][j];
         many=0.0;
         for(i=0;i<m;i++) {
            if(i!=a) {
               many+=y[i][j];
            }
         }
         split[j]= split[j]*one/q[j]*many/q[j];
         split[j]=2*sqrt(split[j]);
      }

      result=0.0;
      for(j=0;j<n;j++) {
         result+=q[j]*split[j];
      }
      final+=(base-result/total);
   }   

   free(p);
   free(q);
   free(split);
   return (final/m);
}

float dkm1_1against_max(float ** y, int m, int n) {
   int i, j;
   float tmp;
   float result =0.0;
   float * p;
   float * q;
   float * split;
   float total=0.0;
   float left, right;
   float base=1.0,final=0.0;
   float one,many;
   int a;

   p=malloc(n*sizeof(float));
   q=malloc(n*sizeof(float));  
   split=malloc(n*sizeof(float));
 
   for(a=0;a<m;a++) {
      total=0;
      base =1.0;
      for(j=0;j<n;j++) {
         q[j]=0;
      }
      for(j=0;j<n;j++) {
         for(i=0;i<m;i++) {
            q[j] +=y[i][j];
            total+=y[i][j];
         }
      }
   
      for(j=0;j<n;j++) {
         p[j] =q[j]/total;
      }
   
      tmp=0.0;
      for(j=0;j<n;j++) {
         tmp+=y[a][j]/total;
      }
      base=base*tmp;
      tmp=0.0;
      for(j=0;j<n;j++) {
         for(i=0;i<m;i++) {
            if(i!=a) {
               tmp+=y[i][j]/total;
            }
         }
      }
      base=base*tmp;
      base=2*sqrt(base);

      for(j=0;j<n;j++) {
         split[j]=1.0;
         one=y[a][j];
         many=0.0;
         for(i=0;i<m;i++) {
            if(i!=a) {
               many+=y[i][j];
            }
         }
         split[j]= split[j]*one/q[j]*many/q[j];
         split[j]=2*sqrt(split[j]);
      }

      result=0.0;
      for(j=0;j<n;j++) {
         result+=q[j]*split[j];
      }
      result=result/total;
      if((base-result) > final) {
         final=(base-result);
      }
   }   

   free(p);
   free(q);
   free(split);
   return (final);
}

