/*
  divide_and_conquer.c

  Given a decision tree, we assign training examples to the correct 
  leaves. Then, we can assign the test examples to the corresponding
  leaves. These sub-training and sub-testing sets can then be used with
  a different classifier, etc. and learned indepedently.
*/

#include <unistd.h>
#include "defns.h"
#include "types.h"
#include "extern.h"


int  _dump_item();


/* External, from build.c */
extern int *train_index, *test_index;
extern int nodeid;

int  *assignments=NULL;
FILE *fp=NULL;

/* The node we are presently using */
Tree conquer_node;
/*
  Divide the training set one item at a time into a particular leaf
  subset 
*/
int divide_train(Boolean use_pruning)
{
  int i,j;
  char filename[1024];
  Tree t;
  
  /* First load the training data (should already be in memory) */
  load_training_data();

  /* Pick the appropriate tree */
  if ( use_pruning ) 
    t=pruned_trees[0];
  else
    t=unpruned_trees[0];

  assignments=CALLOC(number_of_examples,sizeof(int));

  /* Clear out all files (we will be appending like mad) */
  for (i=0; i < nodeid; i++) {
    sprintf(filename,"%s-%d.data",filestem,i);
    unlink(filename);
    sprintf(filename,"%s-%d.idata",filestem,i);
    unlink(filename);
  }

  /* For each item, add the data element to the nth dataset */
  for (i=0; i < number_of_examples; i++) {
    Category(Item[i],t);
    assignments[i]=conquer_node->nodeid;
  }
  
  /* Dump the items to the appropriate files */
  fp=NULL;
  for (i=0; i < nodeid; i++) {
    for (j=0; j < number_of_examples;j++) {
      if ( assignments[j] == i )
	_dump_item(j, i);
    }
    if ( fp != NULL ) {
      fclose(fp);
      fp=NULL;
    }
  }
  
  free(assignments);
  return 0;
}


int divide_test()
{
  char filename[1024];
  FILE *fp;
  int i;
  Tree t;

  /* Load test data (should already be loaded) */
  load_test_data();

  if ( use_pruning ) 
    t=pruned_trees[0];
  else
    t=unpruned_trees[0];

  sprintf(filename,"%s.partition", filestem);
  fp=fopen(filename,"w");
  if ( !fp) {
    printf("Cannot open %s\n",filename);
    exit(1);
  }

  /* For each item, add the data element to the nth dataset */
  for (i=0; i < number_of_test_examples; i++) {
    Category(Item[i],t);
    fprintf(fp,"%d\n",conquer_node->nodeid);
  }
  fclose(fp);

  return 0;
}





/* The Item[i] ended up at node nodeid. Store the example in the 
   appropriate output file (either as an index or as an example)
*/
int _dump_item(int i, int nodeid)
{
  int j;
  char filename[1024];
  
  if ( fp == NULL ) {
    if ( use_index_files ) 
      sprintf(filename,"%s-%d.idata", filestem, nodeid);
    else
      sprintf(filename,"%s-%d.data", filestem, nodeid);

    fp=fopen(filename,"a");
    if (!fp) { fprintf(stderr,"Error opening output"); exit(1);}
  }


  /* Special case of index files, just print index */
  if ( use_index_files ) {
     fprintf(fp, "%d\n", train_index[i]);
  } else {
     for (j=0; j < number_of_attributes; j++)  {
       switch (attribute_type[j]) {
       case BINARY:
         printf("Binary attributes not supported yet.\n");
         break;
       case IGNORE:
         break;
       case DISCRETE:
         fprintf(fp,"%s,",AttValName[j][DVal(Item[i],j)]);
         break;
       case CONTINUOUS:
       default:
         fprintf(fp,"%.5f,",(double)CVal(Item[i], j));
         break;
       }
     }
     fprintf(fp,"%s\n", class_name[Class(Item[i])]);
  } 
  


  return 0;
}
