/*************************************************************************/
/*									 */
/*		Global data for C4.5					 */
/*		--------------------					 */
/* Look at globals.c for descriptions! */
/*									 */
/*************************************************************************/


extern  int		MaxAtt,		/* max att number */
			MaxClass,	/* max class number */
			MaxDiscrVal;	/* max discrete values for any att */

extern  int		MaxItem;	/* max data item number */


extern  int             number_of_examples;
extern  int             number_of_attributes;
extern  int             number_of_classes;

extern  INSTANCE        *train;
extern  INSTANCE        *test;
extern  INSTANCE        *validation;
extern  int             number_of_train_examples;
extern  int             number_of_test_examples;
extern  int             number_of_validation_examples;


extern  INSTANCE	*Item;		/* data items */
extern  INSTANCE        *dataset;

extern  int 		*MaxAttVal;	/* number of values for each att */

extern  int             *attribute_type;

extern  char            *filestem;
extern  char		**class_name,	        /* class names */
		  	**attribute_name,	/* att names */
		  	***AttValName;	/* att value names */

extern  Boolean		AllKnown;	/* true if there have been no splits
					   on atts with missing values above
					   the current position in the tree */

extern Boolean use_validation_set;
extern int validation_percent;
extern Boolean use_subtree_raising;
extern Boolean use_error_estimate;
extern Boolean load_validation;
/*************************************************************************/
/*									 */
/*		Global parameters for C4.5				 */
/*		--------------------------				 */
/*									 */
/*************************************************************************/


extern  int             tree_stats;

extern  int		VERBOSITY;	/* verbosity level (0 = none) */
extern  int
			TRIALS,		/* number of trees to be grown */
	    	        RANDSPLITVAL;   /* number of best splits to choose */
extern  Boolean		GAINRATIO,	/* true=gain ratio, false=gain */
			SUBSET,		/* true if subset tests allowed */
			use_windowing,	/* true if windowing turned off */
			PROBTHRESH,	/* true if to use soft thresholds */
                        divide_and_conquer,
	                collapse_tree,
			  PURE_TREES,
                          OUTPUT_LEAFS;
/* Define the maximum allowed depth at which a random decision is made */
extern int max_random_tree_depth;

/* Define the n, in which to randomly choose from the top-n splits */
extern int rand_split_max;

extern Boolean use_pruning, 
  do_predictions, 
  DELETE_TREES, 
  evaluate_test_set, 
  evaluate_training_set,use_random_attribute_split;
extern Boolean use_boosting;
extern float *BoostingExampleWeights;
extern float *BoostingClassifierWeights;

extern int number_of_trees;
/* General flags for all programs */
extern Boolean use_index_files,
               use_example_weights;
extern Boolean hard_limit_min_branch,use_random_subspaces,use_random_forest;
extern int number_of_random_attributes;

extern  int		MINOBJS,	/* minimum items each side of a cut */
			WINDOW,		/* initial window size */
			INCREMENT;	/* max window increment each iteration */
extern  int          max_tree_depth;

extern float confidence_factor;	
extern int bagging_percent;
extern Boolean use_bagging;


extern char *node_type_string[];

extern Boolean *use_attribute;
extern TREE **pruned_trees, **unpruned_trees;



extern Boolean *inbag;
extern Boolean use_oob;
extern Boolean use_gini;

/*-------------------------------------------------------------------
  Function prototypes

  By convention, I generally name private functions beginning with
  an _ and do not include them here.
  -------------------------------------------------------------------*/

int InitialiseWeights();
void Sprout();
float CountItems();
void Swap();
float GreatestValueBelow(int a, float t);
void Quicksort(int Fp,int Lp, int Att, void (*Exchange)());

/* discr.c */
void EvalDiscreteAtt();
void compute_frequencies();
float DiscrKnownBaseInfo(float KnownItems, int MaxVal);
void DiscreteTest();

/* subset.c */
void EvalSubset(int Att, int Fp, int Lp,ItemCount Items);             
void SubsetTest(Tree Node, Attribute Att);

/* contin.c */
void EvalContinuousAtt();
void ContinTest(TREE *Node, int a);

void reset_frequencies();

/* info.c */
float Worth(float ThisInfo, float ThisGain, float MinGain);
float ComputeGain(float BaseInfo, float UnknFrac, int MaxVal, float TotalItems);
float TotalInfo(float V[], int MinVal, int MaxVal);
void PrintDistribution();

/* log.c */
float Log();
/* genlogs.c */
void GenerateLogs();

/* st-thresh.c */
void SoftenThresh(TREE *T);

int which();
void Error();

/* getdata.c */
void load_data();
void init_data();


/* getnames.c */
void GetNames();
char *read_name();
char *discrete_copystring();

/* besttree.c */
void create_trees();
int find_tree_stats(TREE * tree, int * nodes, int * leafs);
void Classify(Description CaseDesc, Tree T, float Weight);
int LeafID(Description CaseDesc, TREE *T, float Weight);
TREE *FormTree();
Tree CopyTree();
Boolean Prune();
void Swap();
int Category();
int GetLeafID(Description CaseDesc, TREE *DecisionTree);
void Evaluate(Boolean CMInfo, int Saved);
int divide_train();
int divide_test();


/* predict.c */
void predict();


int InitialiseTreeData();

/* trees.c */
void PrintTree();
void SaveTree();
int  TreeSize();
void ReleaseTree();
TREE *create_leaf(int nodeid,float *ClassFreq, int NodeClass, float Cases, float Errors);
void PrintConfusionMatrix();

/* build.c */
int Group(int V, int Fp,int Lp,TREE *TestNode);
int group_unknown_values(int Fp, int Lp, TREE *t);

/* stats.h */
float AddErrs(float,float);

/* header.c */
void PrintHeader();

/* best_splits.c */
void initialize_splits();
void update_best_splits();
int choose_split();

/* windowing.c */
int BestTree();

/* shuffle.c */
void shuffle();

/* evaluate.c */
void evaluate_tree();

/* random_subspaces.c */
void initialize_random_subspaces();
void initialize_random_attribute_subsets();

void save_tree();
void save_pruned_trees();
void save_unpruned_trees();

TREE **load_pruned_trees();
TREE **load_unpruned_trees();


/* prune.c */
TREE *prune_tree();

void print_tree();

/* oob.c */
int calculate_oob();
int init_oob();

/* boosting.c */
void init_boosting();
int BoostingUpdateWeights();

void initialize_peel_off();

