#include <stdlib.h>
#include "globals.h"


/*
 *
 * This file contains code to read the '.names' specification file.
 * This allows the user mercy with white space and '|' to comment lines
 *
 */

char Delimiter;

char *read_word(FILE *fp)
{

  int c,ndx=0;
  char word[MAXLINE];
  char * rm_warning=word;

  bzero(word,MAXLINE);
  
  //printf("%c\n",c);

  //if the EOF is reached, there is no next valid next word
  c=fgetc(fp);
  if ( c == EOF ) {
    Delimiter = EOF;
    return NULL;
  }

  //otherwise, proceed to the next delimiting character
  while ( c != ':' && c != ',' && c != '\n' && c != EOF ) {
    //is this a dot character or the end of the line?
    if ( c == '.' ) {
      if ( ( c = fgetc(fp) ) == '\n' || c == ' ' || c == '\t' )
	{
	  break;
	}
      word[ndx++]='.';
    }

    word[ndx++] = c;
    
    if ( c == ' '  || c == '\t' ){
      while ( ( c = fgetc(fp) ) == ' ' || c == '\t' );
    } else {
      c = fgetc(fp);
    }
    
  }

  //remove trailing whitespace from the returned word
  while ( word[ndx-1] == '\n' || word[ndx-1] == ' ' || word[ndx-1] == '\t' )
    {
      ndx--;
    }
  
  word[ndx++] = '\0';
  return rm_warning;
}


void GetData(int mode)
{
  int attr,tmp;
  FILE * datafp;
  char filename[MAXLINE], *word;
  bzero(filename,MAXLINE);

  if(mode==DATA)
    {
      sprintf(filename,"%s.data", filestem);
      datafp=fopen(filename,"r");

      num_train=0;
      Delimiter=',';
      //read in each example
      word=read_word(datafp);
      while(Delimiter!=EOF)
	{
	  if ( num_train % REALLOC_INCR == 0 )
	    {
	      train_attr_list=realloc(train_attr_list, (num_train+REALLOC_INCR)*sizeof(int *));
	      train_class_list=realloc(train_class_list, (num_train+REALLOC_INCR)*sizeof(int ));
	      train_missing_list=realloc(train_missing_list,(num_train+REALLOC_INCR)*sizeof(int *));
	      if(nominal_only==0)
		{
		  train_continuous_list=realloc(train_continuous_list, (num_train+REALLOC_INCR)*sizeof(float *));
		}
	    }
	  //attributes here
	  train_attr_list[num_train]=malloc(num_attrs*sizeof(int));
	  train_missing_list[num_train]=malloc(num_attrs*sizeof(int));
	  if(nominal_only==0)
	    {
	      train_continuous_list[num_train]=malloc(num_attrs*sizeof(float));
	    }
	  for(attr=0;attr<num_attrs;attr++)
	    {
	      if(Delimiter == '\n')
		{
		  printf("Error reading example %d, too few attributes\n",num_train+1);
		  exit(0);
		}
	      if(attr_types[attr] == CONTINUOUS)
		{
		  if(strcmp(word,"?")==0)
		    {
		      train_missing_list[num_train][attr]=1;
		    }
		  else
		    {
		      train_missing_list[num_train][attr]=0;
		      train_continuous_list[num_train][attr]=atof(word);
		    }
		}
	      else if(attr_types[attr] == IGNORE)
		{
		  if(strcmp(word,"?")==0)
		    {
		      train_missing_list[num_train][attr]=1;
		    }
		  else
		    {
		      train_missing_list[num_train][attr]=0;
		      train_continuous_list[num_train][attr]=atof(word);
		    }
		}
	      else
		{
		  if(strcmp(word,"?")==0)
		    {
		      train_missing_list[num_train][attr]=1;
		    }
		  else
		    {
		      tmp=hash(attr_hash[attr],word);
		      //printf("%d\n",tmp);
		      if(tmp==-1)
			{
			  printf("Error reading .data example %d, unknown attribute value \"%s\" for attribute \"%s\"\n",
				 num_train+1,word,attr_name_list[attr]);
			  exit(0);
			}
		      train_missing_list[num_train][attr]=0;
		      train_attr_list[num_train][attr]=tmp;
		    }
		}
	      word=read_word(datafp);
	    }
	  //classes here
	  train_class_list[num_train]=hash(classes_hash,word);
	  word=read_word(datafp);
	  num_train++;
	}
    }
  else
    {
      /* makey surey this worky */
      sprintf(filename,"%s.test", filestem);
      datafp=fopen(filename,"r");

      num_test=0;
      Delimiter=',';
      //read in each example
      word=read_word(datafp);
      while(Delimiter!=EOF)
	{
	  if ( num_test % REALLOC_INCR == 0 )
	    {
	      test_attr_list=realloc(test_attr_list, (num_test+REALLOC_INCR)*sizeof(int *));
	      test_class_list=realloc(test_class_list, (num_test+REALLOC_INCR)*sizeof(int ));
	      test_missing_list=realloc(test_missing_list,(num_test+REALLOC_INCR)*sizeof(int *));
	      if(nominal_only==0)
		{
		  test_continuous_list=realloc(test_continuous_list, (num_test+REALLOC_INCR)*sizeof(float *));
		}
	    }
	  //attributes here
	  test_attr_list[num_test]=malloc(num_attrs*sizeof(int));
	  test_missing_list[num_test]=malloc(num_attrs*sizeof(int));
	  if(nominal_only==0)
	    {
	      test_continuous_list[num_test]=malloc(num_attrs*sizeof(float));
	    }
	  for(attr=0;attr<num_attrs;attr++)
	    {
	      if(Delimiter == '\n')
		{
		  printf("Error reading example %d, too few attributes\n",num_test+1);
		  exit(0);
		}
	      if(attr_types[attr] == CONTINUOUS)
		{
		  if(strcmp(word,"?")==0)
		    {
		      test_missing_list[num_test][attr]=1;
		    }
		  else
		    {
		      test_missing_list[num_test][attr]=0;
		      test_continuous_list[num_test][attr]=atof(word);
		    }
		}
	      else
		{
		  if(strcmp(word,"?")==0)
		    {
		      test_missing_list[num_test][attr]=1;
		    }
		  else
		    {
		      tmp=hash(attr_hash[attr],word);
		      if(tmp==-1)
			{
			  printf("Error reading .test example %d, unknown attribute value \"%s\" for attribute \"%s\"\n",
				 num_test+1,word,attr_name_list[attr]);
			  exit(0);
			}
		      test_missing_list[num_test][attr]=0;
		      test_attr_list[num_test][attr]=tmp;
		    }
		}
	      word=read_word(datafp);
	    }
	  //classes here
	  test_class_list[num_test]=hash(classes_hash,word);
	  word=read_word(datafp);
	  num_test++;
	}
    }
}
