#include <stdlib.h>
#include "globals.h"


/*
 *
 * This file contains code to read the '.names' specification file.
 * This allows the user mercy with white space and '|' to comment lines
 *
 */

char Delimiter;

char *read_name_word(FILE *fp)
{

  int c,ndx=0;
  char word[MAXLINE];
  char * rm_warning;

  rm_warning=word;
  bzero(word,MAXLINE);
  
  //ignore whitespace and comments preceeding the next word
  while ( ( c = fgetc(fp) ) == '|' || c == '\n' || c == ' ' || c == '\t' ) {
    if ( c == '|' )
      {
	c=SkipRestOfLine(fp);
      }
  }

  //if the EOF is reached, there is no next valid next word
  if ( c == EOF ) {
    Delimiter = EOF;
    return NULL;
  }

  //otherwise, proceed to the next delimiting character
  while ( c != ':' && c != ',' && c != '\n' && c != '|' && c != EOF ) {
    //is this a dot character or the end of the line?
    if ( c == '.' ) {
      if ( ( c = fgetc(fp) ) == '|' || c == '\n' || c == ' ' || c == '\t' )
	{
	  break;
	}
      word[ndx++]='.';
    }

    
    word[ndx++] = c;
    //c = fgetc(fp);
    if ( c == ' '  || c == '\t' ){
      while ( ( c = fgetc(fp) ) == ' ' || c == '\t' );
    } else {
      c = fgetc(fp);
    }
  }

  //handle if the word ends in a comment sign
  if ( c == '|' ) 
    {
      c=SkipRestOfLine(fp);
    }
  Delimiter = c;


  //remove trailing whitespace from the returned word
  while ( word[ndx-1] == '\n' || word[ndx-1] == ' ' || word[ndx-1] == '\t' )
    {
      ndx--;
    }
  
  word[ndx++] = '\0';
  return rm_warning;
}


void GetNames()
{
  FILE * namesfp;
  char filename[MAXLINE], *word;
  int a;

  bzero(filename,MAXLINE);
  sprintf(filename,"%s.names", filestem);

  namesfp=fopen(filename,"r");

  num_classes=0,num_attrs=0;

  //first determine the classes
  //construct the classes_list and classes_hash here
  Delimiter=',';
  init(classes_hash);
  while ( Delimiter == ',' ) {
    word=read_name_word(namesfp);
    if ( num_classes % REALLOC_INCR == 0 )
      {
	classes_list=realloc(classes_list, (num_classes+REALLOC_INCR)*sizeof(char *));
      }
    classes_list[num_classes]=strdup(word);
    
    //do class_hash stuff here
    insert_hash(&classes_hash, word, num_classes++);
  }

  //get number of attributes, and malloc upfront
  Delimiter=',';
  while ( Delimiter != EOF )
    {
      word=read_name_word(namesfp);
      if ( Delimiter == EOF )
	{
	  break;
	}
      if ( Delimiter != ':' )
	{
	  printf("Error reading .names file, inproper attribute specification\n");
	  exit(0);
	}
      word=read_name_word(namesfp);
      if(strcmp(word,"continuous")==0)
	{
	}
      else if(strcmp(word,"ignore")==0)
	{
	}
      else
	{
	  //attr_types[num_attrs]=NOMINAL;
	  while( Delimiter == ',' || Delimiter == '\n')
	    {
	      if ( num_attr_vals[num_attrs] % REALLOC_INCR == 0 )
		{
		}
	      if( Delimiter == '\n' )
		{
		  break;
		}
	      word=read_name_word(namesfp);
	    }
	}
      num_attrs++;
    }

  fclose(namesfp);
  namesfp=fopen(filename,"r");


  Delimiter=',';
  //init(classes_hash);
  while ( Delimiter == ',' ) {
    word=read_name_word(namesfp);
  }

  attr_name_list=malloc((num_attrs)*sizeof(char *));
  num_attr_vals=malloc((num_attrs)*sizeof(int));
  attr_hash=malloc((num_attrs)*sizeof(struct node *));
  attr_list=malloc((num_attrs)*sizeof(char **));
  attr_types=malloc((num_attrs)*sizeof(int));

  //build attr_list
  a=0;
  Delimiter=',';
  while ( Delimiter != EOF )
    {
      word=read_name_word(namesfp);
      if ( Delimiter == EOF )
	{
	  break;
	}
      if ( Delimiter != ':' )
	{
	  printf("Error reading .names file, inproper attribute specification\n");
	  exit(0);
	}
      attr_name_list[a]=strdup(word);
      num_attr_vals[a]=0;
      init(attr_hash[a]);
      //check this guy out
      word=read_name_word(namesfp);
      if(strcmp(word,"continuous")==0)
	{
	  attr_types[a]=CONTINUOUS;
	  nominal_only=0;
	}
      else if(strcmp(word,"ignore")==0)
	{
	  attr_types[a]=IGNORE;
	  nominal_only=0;
	}
      else
	{
	  attr_types[a]=NOMINAL;
	  while( Delimiter == ',' || Delimiter == '\n')
	    {
	      if ( num_attr_vals[a] % REALLOC_INCR == 0 )
		{
		  attr_list[a]=realloc(attr_list[a], (num_attr_vals[a]+REALLOC_INCR)*sizeof(char *));
		}
	      attr_list[a][num_attr_vals[a]]=strdup(word);
	      insert_hash(&(attr_hash[a]),word,num_attr_vals[a]);
	      num_attr_vals[a]++;
	      if( Delimiter == '\n' )
		{
		  break;
		}
	      word=read_name_word(namesfp);
	    }
	}
      a++;
    }
}





//function to move file pointer to the next line
char SkipRestOfLine(FILE *fp)
{
  int c;
  while ( (c=fgetc(fp)) != '\n' );
  return c;
}
