#include <stdio.h>
#include <stdlib.h>
#include <float.h>
#include "hash.h"

//definitions
#define REALLOC_INCR 4
#define MAXLINE 1024
#define DATA 0
#define TEST 1
#define NOMINAL 0
#define CONTINUOUS 1
#define IGNORE 2

#define  Epsilon    1E-3f


//dataset properties
extern char * filestem;
extern int num_classes;
extern int num_attrs;
extern int num_train;
extern int num_test;
extern int nominal_only;
extern int collapse_tree;

//options
extern int MINOBJS;
extern int VERBOSITY;
extern int OUTPUT_LEAFS;
extern int LOAD_TEST_SET;
extern int PURE_TREES;
extern int GAINRATIO;
extern int max_tree_depth;

extern int hard_limit_min_branch;

extern int evaluate_test_set;
extern int do_predictions;
extern float confidence_factor;
extern int use_pruning;
extern int tree_stats;
extern int thain;

extern int euclidean_dist;
extern int chord_dist;
extern int chisquaremetric_dist;
extern int chisquaredistance_dist;
extern int speciesprofile_dist;
extern int hellinger_dist;
extern int hellinger_1againstmax_dist;
extern int hellinger_weightoutside_dist;
extern int hellinger_weightinside_dist;
extern int hellinger_weightboth_dist;
extern int hellinger_inv_dist;
extern int infgain_dist;
extern int gini_dist;
extern int dkm_dist;
extern int dkm_avg_dist;
extern int dkm_1againstavg_dist;
extern int dkm_1againstmax_dist;
extern int gini_skewfree_dist;
extern int purity;

typedef struct tree_node {
  int attribute;
  int num_children;
  float split_val;
  struct tree_node ** children;
  int leaf;
  int * class_dist;
  int class;
  int nodeid;
  int errors;
} tree_node;

typedef struct continuous_sort {
  float value;
  int class_ndx;
} continuous_sort;

//global lists/hashes/tree
extern struct node *      classes_hash;
extern char **            classes_list;
extern char **            attr_name_list;
extern char ***           attr_list;
extern struct node **     attr_hash;
extern int *              num_attr_vals;
extern int *              attr_types;
extern int **             train_attr_list;
extern float **          train_continuous_list;
extern int **             test_attr_list;
extern float **          test_continuous_list;
extern int *              train_class_list;
extern int **             train_missing_list;
extern int *              test_class_list;
extern int **             test_missing_list;
extern struct tree_node * tree_root;
extern int *              num_splits;
extern float **          splits_list;


//functions
int compare (const void * a, const void * b);

//createtree.c
void create_tree();
int find_tree_stats(struct tree_node * tree, int * nodes, int * leafs); 
void make_tree_ID3(struct tree_node **, int *, int *, int *);
//int best_distance_ID3(int *, int *);
int best_ID3(int *, int *);
void make_tree_C45(struct tree_node **, int *, int *, int *, int);
//int best_C45(int *, int *, float *);
int best_C45(int * used_examples, int * used_attr, float * split_val, float * inf, float * avg_distance);
void calc_ID3_2(struct tree_node **, int *, int *);
int compare (const void * a, const void * b);
float Worth(float ThisInfo, float ThisGain, float MinGain);
float calcdistance(int ** N, int m, int n, int * used_examples);

//evaluate.h
void evaluate(int);
void print_class(int);

//getnames.c
char *read_name_word(FILE *fp);
void GetNames();
char SkipRestOfLine(FILE *fp);

//getdata.c
char *read_word(FILE *fp);
void GetData(int);

//splits.c
void initialize_splits();
int  find_splits(float * X, int * Y, int n, float ** S, int * uniques);
