/*
  oob.c

  Handles oob error estimates. Briefly, if we use bagging then the examples that are
  left out of the bag are out-of-bag (oob). We can use the tree to test on these 
  unseen examples as a way of estimating error.

  We can also use this in an ensemble environment by keeping track of the votes for
  each example, when it is considered oob
*/
#include "defns.h"
#include "types.h"
#include "extern.h"


int _majority_vote(int *v);

Boolean *inbag=NULL;
int    **oob_votes=NULL;

int init_oob()
{
  int i;

  if ( inbag != NULL ) free(inbag);
  inbag=CALLOC(number_of_train_examples,sizeof(Boolean));

  if ( oob_votes != NULL ) free(oob_votes);
  oob_votes=CALLOC(number_of_train_examples,sizeof(int *));
  for (i=0; i < number_of_train_examples; i++) {
    oob_votes[i]=CALLOC(number_of_classes, sizeof(int));
  }

  return 0;
}


int calculate_oob(Tree *current_tree)
{
  int i,prediction;
  int num_correct=0,num_used=0;
  int num_vote_correct=0;
  INSTANCE p;
  int mj;

  if ( ! use_bagging ) return -1;

  for (i=0; i < number_of_train_examples; i++) {
    
    /* Get the example */
    p=train[i];


    /* This tests example on current tree and calculates current oob estimate */
    if ( ! inbag[i] ) {
      prediction=Category(p, current_tree);
      if ( Class(p) == prediction ) num_correct++;
      num_used++;
      oob_votes[i][prediction]++;
    }
    
    /* This occurs regardless of current example */
    mj=_majority_vote(oob_votes[i]);
    if ( mj >= 0 && Class(p) == mj ) num_vote_correct++;
  }


  info("OOB Accuracy: bag %5.2f (%d/%d)\t total %5.2f (%d/%d)\n", 
	 100.0*num_correct/(double)num_used, num_correct,num_used,
	 100.0*num_vote_correct/(double)number_of_train_examples,
	 num_vote_correct, number_of_train_examples);


  /* This would be a good point to change the number_of_trees variable to short-circuit
     out of further trees given a low change in oob error estimate (for ensemble) */

  return 0;
}



int _majority_vote(int *v)
{
  int i,max;

  /* Find the max vote */
  for (i=max=0; i < number_of_classes; i++) 
    if ( v[i] > v[max] ) max=i;

  /* If the max vote is zero, no one has voted */
  if ( v[max]==0 ) return -1;

  return max;
}
