#include "globals.h"

void load_into_sort(struct continuous_sort * a, int n, float * X, int * Y);
void load_train_into_sort(struct continuous_sort * a, int attr);
int squish(float * X, int * Y, int * N, int n, float min);

void initialize_global_splits()
{
  int i,j,n,uniques=0;
  float * splits, * X;

  splits_list  = malloc(num_attrs*sizeof(float *));
  num_splits   = malloc(num_attrs*sizeof(int));
  X = malloc(num_train*sizeof(float));
  for(i=0;i<num_attrs;i++) {
    if(attr_types[i] == CONTINUOUS) {
      for(j=0;j<num_train;j++) {
	X[j]=train_continuous_list[j][i];
      }

      n=find_splits(X, train_class_list, num_train, &splits, &uniques);
      if(n>0) {
	splits_list[i]=malloc(n*sizeof(float));
	num_splits[i]=n;
	for(j=0;j<n;j++) {
	  splits_list[i][j]=splits[j];
	}
	free(splits);
      }
    }
  }
  free(X);
}


/* X is a list of continous features and Y a list of their
 respective class indexes. n is the number of entries in
 lists X and Y. Write to S the split values and return
 the number of splits */
int find_splits(float * X, int * Y, int n, float ** S, int * uniques)
{

  /* mallocs and stuff here */
  int i,l,last,sameclass,n_splits = 0, * class_splits, * val_pop;
  float * temp_splits, * vals_list, min_number_in_branch;
  struct continuous_sort * sort_list;

  if(n<MINOBJS) {
    return 0;
  }

  sort_list    = malloc(n*sizeof(struct continuous_sort));
  vals_list    = malloc(n*sizeof(float));
  temp_splits  = malloc(n*sizeof(float));
  class_splits = malloc(n*sizeof(int));
  val_pop      = malloc(n*sizeof(int));

  for(i=0;i<n;i++) {
    vals_list[i]    =0.0;
    temp_splits[i]  =0.0;
    class_splits[i] =0.0;
    val_pop[i]      =0;
  }

  load_into_sort(sort_list, n, X, Y);
  qsort (sort_list, n, sizeof(struct continuous_sort), compare);

  /* 040407 add in */
  if (PURE_TREES)
     min_number_in_branch=MINOBJS;
  else
  {
    min_number_in_branch=0.10 * n / num_classes;
    if ( min_number_in_branch < MINOBJS ) 
      min_number_in_branch = MINOBJS;
    else if ( hard_limit_min_branch == 1 && min_number_in_branch > 25 ) {
      min_number_in_branch = 25;
    }
  }

  sameclass = 1;

  last=0,l=0;
  val_pop[l]++;
  class_splits[l]  = sort_list[0].class_ndx;
  vals_list[l]     = sort_list[0].value;
  (*uniques)+=1;
  for(i=1;i<n;i++) {
    if( (i > min_number_in_branch) && (i < n - min_number_in_branch) ) {
      if ( sort_list[i].value - sort_list[i-1].value > 1E-5 ) {
	(*uniques)+=1;
      }
    }

    if( sort_list[last].value != sort_list[i].value ) {
      vals_list[++l]   = sort_list[i].value;
      class_splits[l]  = sort_list[i].class_ndx;
      last = i;
    }
    else if (class_splits[l] != sort_list[i].class_ndx){
      class_splits[l]=-1;
    }
    val_pop[l]++;
  }
  l++;

  l=squish(vals_list,class_splits,val_pop,l,min_number_in_branch);

  for(i=0;i<l-1;i++) {
    if(class_splits[i] == -1 || class_splits[i+1] == -1 || class_splits[i] != class_splits[i+1]) {
      //Quinlin suggests using the Att Value, not the midpoint, since the
      //midpoint might not occur naturally or meaningfully within the data 
      //temp_splits[n_splits++] = (vals_list[i] + vals_list[i+1])/2.0;
      temp_splits[n_splits++] = vals_list[i];
    }
  }

  if(n_splits>0) {
    (*S) = malloc(n_splits*sizeof(float));
    for(i=0;i<n_splits;i++) {
      (*S)[i] = temp_splits[i];
    }
  }

  /* free the mallocs */
  free(vals_list);
  free(temp_splits);
  free(class_splits);
  free(val_pop);
  free(sort_list);

  return n_splits;
}

void load_into_sort(struct continuous_sort * a, int n, float * X, int * Y)
{
  int i;
  for(i=0;i<n;i++) {
    a[i].value     = X[i];
    a[i].class_ndx = Y[i];
  }
}

void load_train_into_sort(struct continuous_sort * a, int attr)
{
  int i;
  for(i=0;i<num_train;i++)
    {
      a[i].value     = train_continuous_list[i][attr];
      a[i].class_ndx = train_class_list[i];
    }
}

int squish(float * X, int * Y, int * N, int n, float min)
{
  int i,j;
  while(N[0]<min && n > 1) {
     X[0]=X[1];
     if(Y[0] == -1 || Y[1] == -1 || Y[0] != Y[1]) {
	Y[0]=-1;
      }
     N[0]+=N[1];

    for(i=1;i<n-1;i++) {
      X[i]=X[i+1];
      Y[i]=Y[i+1];
      N[i]=N[i+1];
    }
    n--;
  }
  while(N[n-1]<min && n > 1) {
    if(Y[n-2] == -1 || Y[n-1] == -1 || Y[n-2] != Y[n-1]) {
      Y[n-2]=-1;
    }
    N[n-2]+=N[n-1];
    n--;
  }
  return n;
}
