/*************************************************************************/
/*									 */
/*		Type definitions for C4.5				 */
/*              -------------------------				 */
/*									 */
/*************************************************************************/

#include "utils.h"

typedef  char	*Set;

typedef  float	ItemCount;	/* count of (partial) items */

typedef int         DiscrValue;
typedef  int	Attribute;	/* attribute number, 0..MaxAtt */

typedef  union _attribute_value {
  int	_discr_val;
  float	_cont_val;
  int   txt_index;
  float real_value;
} ATTRIBUTE, AttValue, *Description, *INSTANCE;

#define  CVal(Case,Attribute)   Case[Attribute]._cont_val
#define  DVal(Case,Attribute)   Case[Attribute]._discr_val
#define  Class(Case)		Case[MaxAtt+1]._discr_val

#define  Unknown  -999		/* unknown value for continuous attribute */

#define  BrDiscr	1	/* node types:	branch */
#define  ThreshContin	2	/*		threshold cut */
#define  BrSubset	3	/*		subset test */

typedef struct _tree_record TREE;
typedef  struct _tree_record *Tree;
struct _tree_record {
  int     nodeid;
  int	  node_type;	        /* 0=leaf 1=branch 2=cut 3=subset */
  int 	  leaf;		        /* most frequent class at this node */
  float   Items;		/* no of items at this node */
  float   *class_distribution;	/* class distribution of items */
  float   Errors;		        /* no of errors at this node */
  int  	  tested_attribute; 	/* attribute referenced in test */
  int     height;
  float	Cut,		/* threshold for continuous attribute */
    Lower,		/* lower limit of soft threshold */
    Upper;		/* upper limit ditto */
  Set         *Subset;	/* subsets of discrete values  */
  int         num_branches;    /* number of branches at this node */
  TREE	    **branch;	       /* Branch[x] = (sub)tree for outcome x */
};


#define  IGNORE		1	/* special attribute status: do not use */
#define  DISCRETE	2	/* ditto: collect values as data read */
#define  BINARY         3
#define  CONTINUOUS     4

/* 0=leaf 1=branch 2=cut 3=subset */
#define LEAF_NODE 0
#define DISCRETE_NODE 1
#define CONTINUOUS_NODE 2
#define SUBSET_NODE 3



#define UNPRUNED_TREE     1
#define PRUNED_TREE       2

typedef short	RuleNo;			/* rule number */


typedef struct TestRec *Test;

struct TestRec
       {
	 int  node_type;	/* test type (see tree nodes) */
	 int  tested_attribute;		/* attribute tested */
	 int  num_branches;		/* possible branches */
	 float	Cut;		/* threshold (if relevant) */
	 Set		*Subset;	/* subset (if relevant) */
       };


typedef struct CondRec *Condition;

struct CondRec
       {
	   Test		CondTest;	/* test part of condition */
	   short	TestValue;	/* specified outcome of test */
       };


typedef struct ProdRuleRec PR;

struct ProdRuleRec
       {
	   short	Size;		/* number of conditions */
	   Condition	*Lhs;		/* conditions themselves */
	   int		Rhs;		/* class given by rule */
	   float	Error,		/* estimated error rate */
			Bits;		/* bits to encode rule */
	   int		Used,		/* times rule used */
			Incorrect;	/* times rule incorrect */
       };


typedef struct RuleSetRec RuleSet;

struct RuleSetRec
       {
	   PR		*SRule;		/* rules */
	   RuleNo	SNRules,	/* number of rules */
			*SRuleIndex;	/* ranking of rules */
	   int		SDefaultClass;	/* default class for this ruleset */
    };



typedef struct SplitList_Struct SplitList;
struct SplitList_Struct {
  /* Number of splits so far */
  int    number_of_splits;

  /* Index of attribute for the split */
  int *attribute;

  /* Statistics for the split */
  float *gain;
  float *info;
  float *gain_ratio;

  /* For continuous attributes, at least, the threshold */
  float *bar;
};
