/*
   bits.c

   Handles all of the binary attribute issues. They include reading in a
   binary string (in hex form) and converting to the internal 
   representation, which is an array of unsigned char's. 
*/
#include <stdio.h>
#include <stdlib.h>
#include "utils.h"

/* Each bit position */
BYTE setbits[8]={128,64,32,16,8,4,2,1};


/********************************************************************
  Binary feature support
  This library support bit features (0/1) represented as an array
  of 8-bit bytes (unsigned chars). 
     
  There are several methods of working with them. Note that they
  are general not the best to use for large numbers of bits, since
  they are more inefficient than handling one byte at a time. But,
  they are simple to understand and use.
*********************************************************************/

/* numbytes(n) - Given n bits, return number of bytes needed/used */
int numbytes(int n)
{
   register unsigned int i = n;
   register unsigned int d = i>>3;

   if ( (i&0x7) > 0 ) d++;

   return d;
}

/* bget(v,index) - Return the index'th bit in the bitstring v */
int bget(BYTE *v,int index)
{
   register unsigned int i = index;
   register unsigned int o = i & 0x7;
   register unsigned int d = i >> 3;

   return ((v[d] & setbits[o]) > 0);
}


/* bset(v,index,value) - Set the index'th bit in the bitstring v to value */
void bset(BYTE *v,int index,int value)
{
   register unsigned int i = index;
   register unsigned int o = i & 0x07;
   register unsigned int d = i >> 3;
   if ( (value) == 0 )
      v[d] &= (BYTE)~setbits[o];
   else
      v[d] |= setbits[o];

   return;
}


