/* endian.c -- A trick for determining the byte order of a machine. */

/*
 * The conversion code was taken from Nicolas Devillard at ndevilla.free.fr
 * No license posted with code.
 */
 
/* 
 * The endian detection code was taken from the Bash shell and falls
 * under the following license:
 * Copyright (C) 1993 Free Software Foundation, Inc.

   This file is part of GNU Bash, the Bourne Again SHell.

   Bash is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.

   Bash is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.

   You should have received a copy of the GNU General Public License along
   with Bash; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */

#include <sys/types.h>
#include <stdio.h>



#ifndef INTERROGATE_ENDIAN

/* Declare the endian nature as a variable (as well as define) */
#include "utils.h"
#ifdef BIG_ENDIAN
 int endian=SYS_BIG_ENDIAN;
#else
 int endian=SYS_LITTLE_ENDIAN;
#endif

int swap_int(unsigned int dw)
{
    register int tmp;
    tmp =  (dw & (unsigned int)0x000000FF);
    tmp = ((dw & (unsigned int)0x0000FF00) >> 0x08) | (tmp << (unsigned int)0x08);
    tmp = ((dw & (unsigned int)0x00FF0000) >> 0x10) | (tmp << (unsigned int)0x08);
    tmp = ((dw & (unsigned int)0xFF000000) >> 0x18) | (tmp << (unsigned int)0x08);
    return tmp;
}

short swap_short(unsigned short w)
{
  register short tmp;
  tmp=(w & (unsigned short)0x00FF);
  tmp=((w & (unsigned short)0xFF00) >> (unsigned short)0x08) | ( tmp << (unsigned short)0x08);

  return tmp;
}

#else

char nstring[9];

/* Stuff "1234" into a long, and compare it against a character string
   "1234".  If the results are EQ, the machine is big endian like a 68000
   or Sparc, otherwise it is little endian, like a Vax, or 386. */
main (int argc,char **argv)
{
#if defined (__STDC__)
  register size_t i;
#else
  register int i;
#endif /* !__STDC__ */
  union {
      unsigned long l;
      char s[sizeof (long)];
  } u;
	char string[9];

  if (sizeof (long int) == 4) {
      u.l = 0x04030201L;
      (void) strcpy (string, "4321");
  }
  else if (sizeof (long int) == 8)
  {
    #if defined (__GNUC__)
      unsigned long fake_out_gcc;

      fake_out_gcc = (0x08070605L << 31);
      fake_out_gcc = (fake_out_gcc << 1);
      u.l = fake_out_gcc | 0x04030201L;
    #else
      u.l = (0x08070605L << 32) | 0x04030201L;
    #endif /* !__GNUC__ */
    (void) strcpy (string, "87654321");
  }
  else
  {
   fprintf (stderr,
	       "sizeof (long int) = %d, which isn't handled here.\n",
	        sizeof (long int));
   exit (2);
    }

  for (i = 0; i < sizeof (long); i++)
    nstring[i] = u.s[i] + '0';
  nstring[i] = '\0';
	printf("#ifndef __CONFIG_H\n#define __CONFIG_H\n");
	if (strcmp (nstring, string) == 0)
		printf("#ifndef BIG_ENDIAN\n #define BIG_ENDIAN\n#endif\n");
	else
		printf("#ifndef LITTLE_ENDIAN\n #define LITTLE_ENDIAN\n#endif\n");
	printf("#define SYS_BIG_ENDIAN 1\n#define SYS_LITTLE_ENDIAN 2\n#endif\n");
	exit (0);
}
#endif
