#ifndef __UTIL_LOG_H__
#define __UTIL_LOG_H__
/*********************************************************************
 * 
 * Logging utilities.
 * The method used is to define generic macros that capture some extra
 * information, etc. and then pass them off to actual implementations.
 * These implementations are part of separate libraries (libtxtlog and
 * libwxlog), so that different types of implementations are provided.
 *********************************************************************/


/* 
 * Define the implementation of each function. They must be
 * defined as C functions if compiling using C++.
 */
#ifdef __cplusplus
extern "C" {
#endif
	/* Logging utilities */
	void die_impl(char *msg,...);
	void debug_impl(char *msg, ...);
	void warn_impl(char *msg, ...);
	void info_impl(const char *msg,...);
	void status_impl(const char *msg,...);

	/* Progress bar utilities */
	int pb_create(int max,char *title, ...);
	void pb_update(int, int, char *,...);
	void pb_finish(int);

#ifdef __cplusplus
};
#endif




 /* info() - */
 #define info(msg,args...)  \
 			({ \
 				info_impl(msg,##args); \
 			})
 #define status(msg,args...) \
 			({ \
 				status_impl(msg,##args); \
 			})

      	
 /* warn() - info with source file/line number */
 #define warn(msg,args...)  \
				warn_impl("%s:%d " msg "\n",__FILE__,__LINE__, ##args)
             

 /* debug(level, printf). Output msg when debug_level >= level */
 #define debug(level,msg,args...) \
    do { \
       if ( debug_level >= level ) \
				debug_impl("%s:%d " msg "\n",__FILE__,__LINE__, ##args)\
    } while (0)

 /* die() - a perlism, printf macro that also dies */
 #define die(msg,...) \
			die_impl("%s:%d " msg "\n",__FILE__,__LINE__, ##__VA_ARGS__)
/* 
 * Yield control of processing. Mostly useful if in a GUI, so that other
 * events can occur. Particularly, processing can be interrupted this way.
 */
 #define yield() \
        ({ \
        	 return yield_impl; \
        })


#endif

